# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import builtins
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Set, Tuple, Type, TypeVar, Union

from rich.tree import Tree

from chipscopy.api._detail.property import PropertyCommands, Watchlist, WatchlistEventHandler
from chipscopy.api.containers import QueryList
from chipscopy.client.ibert_core_client import IBERTCoreClient
from typing_extensions import Final

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.api.ibert import IBERT


class IBERTWatchlist(Watchlist["IBERTPropertyCommands"]):
    def __init__(self, parent):
        Watchlist.__init__(self, parent)

    def add(
        self,
        property_names: Union[str, List[str]],
        *,
        listeners: Union[WatchlistEventHandler, List[WatchlistEventHandler]] = None,
    ):
        """
        Add properties to the watchlist

        Args:
            property_names: Property name(s)
            listeners: Callback to be called when the property update event is received from cs_server

        """
        if listeners is not None and not isinstance(listeners, list):
            listeners = [listeners]

        sanitized_data = PropertyCommands.sanitize_input(property_names, list)

        self.core_tcf_node.add_to_property_watchlist(sanitized_data, self.parent.endpoint_name)

        if not self._tcf_event_listener_registered:
            self.parent.endpoint_tcf_node.add_listener(self._mandatory_event_listener)
            self._tcf_event_listener_registered = True

        watch_id = self._get_uuid()
        self._listeners[watch_id].extend(listeners)

        for property_name in sanitized_data:
            self._watch_ids_for_property[property_name].add(watch_id)

    def remove(self, property_names: Union[str, List[str]]):
        """
        Remove properties from watchlist

        Args:
            property_names: Property name(s)

        """
        sanitized_data = {
            prop
            for prop in PropertyCommands.sanitize_input(property_names, list)
            if prop in self._watch_ids_for_property
        }

        self.core_tcf_node.remove_from_property_watchlist(sanitized_data, self.parent.endpoint_name)

        for prop in sanitized_data:
            del self._watch_ids_for_property[prop]


class IBERTPropertyCommands(PropertyCommands["SerialObjectBase"]):
    def __init__(self, parent, property_endpoint):
        PropertyCommands.__init__(self, parent)

        self.watchlist = IBERTWatchlist(self)
        self.endpoint_name: Final[str] = property_endpoint.name
        self.endpoint_tcf_node = self.core_tcf_node.get_child_with_name(self.endpoint_name)

    """
    NOTE - The 'endpoint_tcf_node' is different from 'core_tcf_node'.
    core_tcf_node - Points to the IBERTCoreClient class and is needed to call server commands
    _endpoint_tcf_node - Points to the Node class which is a child of core_tcf_node. 
     Needed to access props attached to a node
     
    Example with IBERT hierarchy 
        IBERT Versal GTY  <-- core_tcf_node
        ├── Quad_206 <-- if self.endpoint_name == "Quad_206", endpoint_tcf_node points to this
        ├── Quad_204 <-- if self.endpoint_name == "Quad_204", endpoint_tcf_node points to this 
        ├── Quad_205 <-- if self.endpoint_name == "Quad_205", endpoint_tcf_node points to this 
        ├── Quad_203 <-- if self.endpoint_name == "Quad_203", endpoint_tcf_node points to this 
        ├── Quad_202 <-- if self.endpoint_name == "Quad_202", endpoint_tcf_node points to this  
        └── Quad_201 <-- if self.endpoint_name == "Quad_201", endpoint_tcf_node points to this
    """

    def _segregate_props(self, all_property_names) -> Tuple[Set[str], Set[str]]:
        all_property_names = set(all_property_names)

        client_side = set(self.endpoint_tcf_node.props.keys()).intersection(all_property_names)
        # Only use properties that are still in the watchlist. If a prop is deleted in the server
        # TCF node, it wont get deleted in the client TCF node. The prop value in that case would
        # be stale, since it's no longer updated via TCF property update events from the server.
        client_side.intersection_update(self.watchlist.active_properties)

        server_side = all_property_names.difference(client_side)
        return client_side, server_side

    def get(self, property_names: Union[str, List[str]]) -> Dict[str, Any]:
        """
        Get the property value cached in cs_server

        Args:
            property_names: Property name(s)

        Returns:
            Dictionary with property name and value as key, value pairs.

        """
        sanitized_data = PropertyCommands.sanitize_input(property_names, list)
        client_side_props, server_side_props = self._segregate_props(sanitized_data)

        all_property_values = dict()

        # Fetch client side properties from Node.props
        for property_name in client_side_props:
            all_property_values[property_name] = self.endpoint_tcf_node.props[property_name]

        if len(server_side_props) > 0:
            # If there are properties not available with the client, go to the server to fetch them
            all_property_values.update(
                self.core_tcf_node.get_property(list(server_side_props), self.endpoint_name)
            )

        return all_property_values

    def set(self, **property_dict):
        """
        Set new values for properties in cs_server

        Args:
            **property_dict:
             Unpacked dict with key as property and value as new property value

        """
        sanitized_data = PropertyCommands.sanitize_input(property_dict, dict)
        return self.core_tcf_node.set_property(sanitized_data, self.endpoint_name)

    def refresh(self, property_names: Union[str, List[str]]) -> Dict[str, Any]:
        """
        Refresh the value of properties and update it in cs_server

        Args:
            property_names: Property name(s)

        Returns:
            Dictionary with property name and refreshed value as key, value pairs.

        """
        sanitized_data = PropertyCommands.sanitize_input(property_names, list)
        return self.core_tcf_node.refresh_property(sanitized_data, self.endpoint_name)

    def commit(self, property_names: Union[str, List[str]]):
        """
        Commit the value of properties to HW

        Args:
            property_names: Property name(s)

        """
        sanitized_data = PropertyCommands.sanitize_input(property_names, list)
        return self.core_tcf_node.commit_property(sanitized_data, self.endpoint_name)

    def report(self, property_names: Union[str, List[str]] = None) -> Dict[str, Dict[str, Any]]:
        """
        Generate a report providing detailed information about the properties.

        Args:
            property_names: Property name(s)

        Returns:
            Dictionary with property name and information as key, value pairs.

        """
        sanitized_data = []
        if property_names is not None:
            sanitized_data = PropertyCommands.sanitize_input(property_names, list)
        return self.core_tcf_node.report_property(sanitized_data, self.endpoint_name)


parent_type = TypeVar("parent_type")
child_type = TypeVar("child_type")


class SerialObjectBase(Generic[parent_type, child_type]):
    """
    Abstract base class for all serial objects
    """

    def __init__(
        self,
        *,
        name: str,
        type: str,
        parent: parent_type,
        handle: str,
        core_tcf_node: IBERTCoreClient,
        property_for_alias: Dict[str, str],
        is_property_endpoint: bool,
        modifiable_aliases: Set[str],
    ):
        self.name: Final[str] = name
        """Object name"""

        self.type: Final[str] = type
        """Object type"""

        self.handle: Final[str] = handle
        """Object handle from cs_server"""

        self.parent: Final[parent_type] = parent
        """Parent of this object"""

        self.children: QueryList[child_type] = QueryList()
        """Direct children of this object"""

        self.core_tcf_node: Final[IBERTCoreClient] = core_tcf_node

        self.property_for_alias: Dict[str, str] = property_for_alias
        """Alias to property name mapping"""

        self.modifiable_aliases: Set[str] = modifiable_aliases

        self._property_endpoint: Type[SerialObjectBase]
        if is_property_endpoint:
            self._property_endpoint = self
        else:
            if self.parent._property_endpoint is None:
                raise ValueError(
                    f"Property endpoint for {self.parent.handle} is None! "
                    f"Cannot determine property endpoint for {self.handle}"
                )

            self._property_endpoint = self.parent._property_endpoint

        self.property: IBERTPropertyCommands = IBERTPropertyCommands(
            parent=self,
            property_endpoint=self._property_endpoint,
        )

        self.filter_by = {}

    def __repr__(self):
        return f"{self.type} - {self.name} - {self.handle}"

    def __rich_tree__(self):
        root = Tree(self.name)
        for child in self.children:
            try:
                root.add(child.__rich_tree__())
            except AttributeError:
                root.add(child.name)

        return root

    # Explicitly use property from builtins here, since this class has instance var with same name
    @builtins.property
    def aliases(self) -> Set[str]:
        """

        Returns:

        """
        return set(self.property_for_alias.keys())

    def reset(self):
        raise NotImplementedError(f"Reset is not supported for {self.handle}!")



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

from .. import test


class TestProxy(test.TestService):
    def add_manager(self, manager_name, done):
        return self.send_xicom_command("addManager", (manager_name,), done)

    def remove_manager(self, manager_name, done):
        return self.send_xicom_command("removeManager", (manager_name,), done)

    def add_node(self, manager_name, ctx, props, done):
        return self.send_xicom_command("addNode", (manager_name, ctx, props), done)

    def remove_node(self, manager_name, ctx, done):
        return self.send_xicom_command("removeNode", (manager_name, ctx), done)

    def update_node(self, manager_name, ctx, props, done):
        return self.send_xicom_command("updateNode", (manager_name, ctx, props), done)

    def exit(self, done):
        return self.send_xicom_command("exit", (), done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from chipscopy.tcf.channel.Command import Command
from chipscopy.tcf.services import Service, DoneHWCommand, Token

JTAG_CABLE_SERVICE = "JtagCable"


class JtagCableProxy(Service):
    def __init__(self, channel):
        self.channel = channel
        self.listeners = {}

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return JTAG_CABLE_SERVICE

    def getServerDescriptions(self, done):
        done = self._makeCallback(done)
        service = self

        class GetServerDescriptionsCommand(Command):
            def __init__(self):
                super(GetServerDescriptionsCommand, self).__init__(
                    service.channel, service, "getServerDescriptions", None
                )

            def done(self, error, args):
                argResult = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    argResult = args[1]
                done.doneHW(self.token, error, argResult)

        return GetServerDescriptionsCommand().token

    def getOpenServers(self, done):
        done = self._makeCallback(done)
        service = self

        class GetOpenServersCommand(Command):
            def __init__(self):
                super(GetOpenServersCommand, self).__init__(
                    service.channel, service, "getOpenServers", None
                )

            def done(self, error, args):
                argResult = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    argResult = args[1]
                done.doneHW(self.token, error, argResult)

        return GetOpenServersCommand().token

    def getServerContext(self, server_id, done):
        done = self._makeCallback(done)
        service = self

        class GetServerContextCommand(Command):
            def __init__(self):
                super(GetServerContextCommand, self).__init__(
                    service.channel, service, "getServerContext", (server_id,)
                )

            def done(self, error, args):
                argResult = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    argResult = args[1]
                done.doneHW(self.token, error, argResult)

        return GetServerContextCommand().token

    def getPortDescriptions(self, server_id, done):
        done = self._makeCallback(done)
        service = self

        class GetPortDescriptionsCommand(Command):
            def __init__(self):
                super(GetPortDescriptionsCommand, self).__init__(
                    service.channel, service, "getPortDescriptions", (server_id,)
                )

            def done(self, error, args):
                argResult = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    argResult = args[1]
                done.doneHW(self.token, error, argResult)

        return GetPortDescriptionsCommand().token

    def getContext(self, port_id, done):
        done = self._makeCallback(done)
        service = self

        class GetPortContextCommand(Command):
            def __init__(self):
                super(GetPortContextCommand, self).__init__(
                    service.channel, service, "getContext", (port_id,)
                )

            def done(self, error, args):
                argResult = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    argResult = args[1]
                done.doneHW(self.token, error, argResult)

        return GetPortContextCommand().token

    def openServer(self, server_id, params, done):
        return self.send_xicom_command("openServer", (server_id, params), done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from inspect import getattr_static
from typing import TypeVar, Sequence, Any
from collections import UserDict, UserList

# This along with the Generic[T] base class for QueryList is needed to allow type hints such as
# QueryList[VIO] - Implies the QueryList contains only VIO cores
T = TypeVar("T")


# The Generic[T] adds the ability to subscript i.e. use the [], for QueryList in type hints
class QueryList(UserList, Sequence[T]):
    """
    QueryList is a simple custom list with a few extra bells and whistles.

    QueryList elements can be any object. If the elements contain a dictionary member called "filter_by",
    those key/value pairs in the dictionary are used for filtering.

    A custom matching function can be specified in this querylist using set_custom_match_function. Any
    custom match function takes priority when matching over the optional filter_by attribute.
    """

    def __init__(self, initlist=None):
        self.custom_match_function = None
        super().__init__(initlist=initlist)

    def set_custom_match_function(self, match_function):
        """Sets an optional custom match function for the QueryList. If a custom match function is set,
        each list item is compared using the match function. The match function is expected to return
        True if the list_item matches the passed in key/value pair.

        Args:
            match_function(list_item, key, value):: A custom list element matching function
        """
        self.custom_match_function = match_function

    def all(self) -> "QueryList[T]":
        """
        Returns: every element in list
        """
        return self

    def get(self, **filters: Any) -> T:
        """Get one list item matching the filter. If exactly one list item does not match the filter, a ValueError
        is raised.

        Args:
            **filters ():

        Returns:
            The one element matching the filter
        """
        if filters:
            filtered_result = self.filter_by(**filters)
        else:
            filtered_result = self.all()

        if len(filtered_result) != 1:
            raise ValueError(
                f"Error:QueryList:get() - returned {len(filtered_result)} objects instead of exactly 1"
            )
        return filtered_result[0]

    def at(self, index) -> T:
        """
        Return value from index

        Args:
            index: Index of the element to get

        Returns:
            Element at index
        """
        return self[index]

    # TODO: Consider using filter() instead of filter_by() for the function name. Discuss with dev group.
    # def filter(self, **filters: Any) -> "QueryList[T]":
    #     """Same as filter_by. Not sure which function name to use yet. See documentation for filter_by.
    #     """
    #     return self.filter_by(**filters)

    def filter_by(self, **filters: Any) -> "QueryList[T]":
        """
        Iterate through list and return only those elements that match specifications in filters.
        For an item to qualify as 'filterable', the item must be an object and have an attribute
        named `filter_by` of type Dict[str, Any].

        This function will use the filters provided to it and match all the specifications
        against the data in the `filter_by` attribute of the each element in the list.

        Args:
            **filters:

        Returns:
            List of elements that match the specifications provided
        """
        retval = QueryList()
        retval.set_custom_match_function(self.custom_match_function)

        for list_item in self:
            # First priority is a custom match function set on this QueryList.
            # Custom match function works across all list elements without regard to filter_by property.
            if self.custom_match_function:
                # Non-default match function - this is the first priority to check for a match
                matched = True
                for key, value in filters.items():
                    if not self.custom_match_function(list_item, key, value):
                        matched = False
                if matched:
                    retval.append(list_item)
            else:
                # To support filtering the object must have 'filter_by' attribute as a second priority to
                # the custom match function.
                # filter_by is a dict where the key is a property name, and value is a property value.
                if not hasattr(list_item, "filter_by"):
                    continue
                matched = True
                for key, value in filters.items():
                    # If the object has it's own staticmethod 'check_for_filter_match()', call it
                    # else use the static 'check_for_filter_match' in this class
                    match_function = QueryList.check_for_filter_match
                    try:
                        _ = getattr_static(list_item, "check_for_filter_match")
                        match_function = list_item.check_for_filter_match
                    except AttributeError:
                        pass
                    if not match_function(list_item.filter_by, key, value):
                        matched = False
                if matched:
                    retval.append(list_item)

        return retval

    @staticmethod
    def check_for_filter_match(filters_dict, filter_name, match_value) -> bool:
        if filter_name not in filters_dict:
            return False

        pattern = str(match_value)
        string_to_match = str(filters_dict[filter_name])
        match_result = re.fullmatch(pattern, string_to_match)
        return match_result is not None


class QueryDict(UserDict):
    """
    Dictionary with bells and whistles
    """

    def flatten(self, sort_by_key: bool = True, separator: str = "/"):
        """
        Flatten a nested dictionary

        Args:
            sort_by_key: True if keys should be sorted alphabetically

            separator: The character(s) to use when joining parent and child key

        """

        def flattener(input_dict):
            return_val = dict()
            for key, value in input_dict.items():
                if not isinstance(value, dict):
                    return_val[key] = value
                else:
                    for k, v in value.items():
                        if not isinstance(v, dict):
                            return_val[f"{key + separator + k}"] = v
                        else:
                            return_val.update(flattener({f"{key + separator + k}": v}))
            return return_val

        x = QueryDict(flattener(self))
        if sort_by_key:
            return {key: x[key] for key in sorted(x.keys())}
        return x



# *****************************************************************************
# * Copyright (c) 2011, 2013, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"Internal utility methods used for logging/tracing."

import locale
import time

from .. import protocol

DECIMAL_DELIMITER = locale.localeconv().get('decimal_point', '.')


def getDebugTime():
    """
    Returns a relative timestamp in the form "seconds,milliseconds". Each
    segment is zero-padded to three digits, ensuring a consistent length of
    seven characters. The timestamp has no absolute meaning. It is merely the
    elapsed time since January 1, 1970 UT truncated at 999 seconds. Do not
    use this for production code, especially for mathematically determining
    the relative time between two events, since the counter will flip to zero
    roughly every 16 minutes.
    """
    traceBuilder = ''

    # Record the time
    tm = int(time.time() * 1000)
    seconds = int((tm / 1000)) % 1000
    if seconds < 100:
        traceBuilder += '0'
    if seconds < 10:
        traceBuilder += '0'
    traceBuilder += str(seconds)
    traceBuilder += str(DECIMAL_DELIMITER)
    millis = tm % 1000
    if millis < 100:
        traceBuilder += '0'
    if millis < 10:
        traceBuilder += '0'
    traceBuilder += str(millis)
    return traceBuilder


def trace(msg):
    """
    Trace hooks should use this method to log a message. It prepends the
    message with a timestamp and sends it to the TCF logger facility. The
    logger implementation may or may not inject its own timestamp. For
    tracing, we definitely need one, so we introduce a minimal, relative-time
    stamp.

    @param msg  the trace message
    """
    protocol.log('%s %s' % (getDebugTime(), msg))



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from abc import abstractmethod
from typing import List, Dict, Any, Union

from chipscopy.client.core import CoreClient
from chipscopy.dm.request import DoneCallback
from chipscopy.utils import listify

PropertyValues = Dict[str, Any]


class PropertyPermission(enum.IntFlag):
    """
        Access allowed for properties.
        Bit field to check what methods are allowed on a particular property.
        See :meth:`AxisILAService.report_property`

        When passing enum values to/from API functions, use int values.

    ::

        COMMIT  = 0x01
        GET     = 0x02
        REFRESH = 0x04
        RESET   = 0x08
        SET     = 0x10
    """

    #: Commit to core
    COMMIT = 0x01

    #: Get from core object
    GET = 0x02

    #: Refresh data in core object with data from the core
    REFRESH = 0x04

    #: Reset the value in the core object
    RESET = 0x08

    #: Set the value in the core object
    SET = 0x10

    @staticmethod
    def get_str_representation(permission_as_int: int) -> List[str]:
        permissions = list()

        permission_as_bin = bin(permission_as_int)[2:].zfill(5)
        if permission_as_bin[-1] == "1":
            permissions.append("COMMIT")

        if permission_as_bin[-2] == "1":
            permissions.append("GET")

        if permission_as_bin[-3] == "1":
            permissions.append("REFRESH")

        if permission_as_bin[-4] == "1":
            permissions.append("RESET")

        if permission_as_bin[-5] == "1":
            permissions.append("SET")

        return permissions


class CorePropertyClient(CoreClient):
    @abstractmethod
    def get_service_proxy(self):
        pass

    def get_property(self, property_names: List[str], done: DoneCallback = None):
        """
        Get values for properties.

        :param property_names: List of property names. Empty list means get all properties.
        :param done: Callback with result and any error.
        :return: Property name/value pairs in a dict. See :data:`PropertyValues`.
        """
        service, done_cb = self.make_done(done)
        token = service.get_property(self.ctx, property_names, done_cb)
        return self.add_pending(token)

    def get_property_group(self, groups: List[str], done: DoneCallback = None):
        """
        Get property values for property groups.

        :param groups: List of property group names. Empty list means all properties.
        :param done: Callback with result and any error.
        :return: Properties for named group names. :data:`PropertyValues`
        """
        service, done_cb = self.make_done(done)
        token = service.get_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    def report_property(self, property_names: List[str], done: DoneCallback = None):
        """
        Return description of properties.

        Example: Description of property 'state'
        ::

            Attributes:

            group_name - Property group which the property belongs to.
            type - Property value type
            permission - Permissions control which methods can operate on property.
            name - Name of property
            default_value - Reset value for property.
            value - Current value of the property.

            {
                'state': {
                    'group_name': 'core_info',
                    'type': int),
                    'permission': 2,        # corresponds to bitwise enum value <_PropertyPermission.GET: 2>,
                    'name': 'core_major_ver',
                    'default_value': 0,
                    'value': 3
                }
            }

        For further information about 'permission', see :class:`_PropertyPermission`

        :param property_names: List of property names. Empty list means get all properties.
        :param done: Callback with result and any error.
        :return: Dict of property description dicts.
        """
        service, done_cb = self.make_done(done)
        token = service.report_property(self.ctx, property_names, done_cb)
        return self.add_pending(token)

    def report_property_group(self, groups: List[str], done: DoneCallback = None):
        """
        Return description of properties, which belong to argument 'groups'.

        :param groups: List of property group names. Empty list means all properties.
        :param done: Callback with result and any error.
        :return: Dict of property description dicts.
        """
        service, done_cb = self.make_done(done)
        token = service.report_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    def reset_property(self, property_names: List[str], done: DoneCallback = None):
        """
        Reset property values, to default values.

        :param property_names: List of property names. Empty list means all resettable properties.
        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.reset_property(self.ctx, property_names, done_cb)
        return self.add_pending(token)

    def reset_property_group(self, groups: List[str], done: DoneCallback = None):
        """

        :param groups:  List of property group names. Empty list means all resettable properties.
        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.reset_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    def set_property(self, property_values: Dict[str, Any], done: DoneCallback = None):
        """
        Set properties. If no properties specified, do nothing.
        Validates properties values for correctness.

        :param property_values: Property name/value pairs.
        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.set_property(self.ctx, property_values, done_cb)
        return self.add_pending(token)

    def set_property_group(self, property_values: Dict[str, Any], done: DoneCallback = None):
        """
        Set properties. If no properties specified, do nothing.
        Validates properties values for correctness.

        :param property_values: Property name/value pairs.
        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.set_property_group(self.ctx, property_values, done_cb)
        return self.add_pending(token)

    def commit_property_group(self, groups: List[str], done: DoneCallback = None):
        """
        Write current property group values to the core.

        :param groups: Group name or list of names.
            Empty list means property groups, which can be written to the core.
        :param done: Callback with result and any error.
        """
        service, done_cb = self.make_done(done)
        token = service.commit_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    def refresh_property_group(self, groups: List[str], done: DoneCallback = None):
        """
        Read values from the core, update ILA properties
        and return those property values to caller.

        :param groups: List of group names.
            Empty list means all refreshable groups.
        :param done: Callback with result and any error.
        :returns: Property name/value pairs.
        """
        service, done_cb = self.make_done(done)
        token = service.refresh_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    def list_property_groups(self, done: DoneCallback = None):
        service, done_cb = self.make_done(done)
        token = service.list_property_groups(self.ctx, done_cb)
        return self.add_pending(token)

    def add_to_property_watchlist(
        self, property_names: Union[str, List[str]], *, done: DoneCallback = None
    ):
        service, done_cb = self.make_done(done)
        property_names = listify(property_names)
        token = service.add_to_property_watchlist(property_names, done_cb)
        return self.add_pending(token)

    def remove_from_property_watchlist(
        self, property_names: Union[str, List[str]], *, done: DoneCallback = None
    ):
        service, done_cb = self.make_done(done)
        property_names = listify(property_names)
        token = service.remove_from_property_watchlist(property_names, done_cb)
        return self.add_pending(token)

    def commit_memory(
        self,
        property_name: str,
        data: bytearray,
        start_byte_index: int,
        word_byte_length: int,
        done: DoneCallback = None,
    ):
        """
        Test function to write to debug core memory

        :param property_name: Name of memory property.
        :param data: Binary data to be written.
        :param start_byte_index:  Memory byte index, to start writing data. Must be on word boundary
        :param word_byte_length: Number of bytes per memory word.
        """
        service, done_cb = self.make_done(done)
        token = service.commit_memory(
            self.ctx, property_name, data, start_byte_index, word_byte_length, done_cb
        )
        return self.add_pending(token)

    def refresh_memory(
        self,
        property_name: str,
        byte_count: int,
        start_byte_index: int,
        word_byte_length: int,
        done: DoneCallback = None,
    ):
        """
        Test function to read debug core memory

        :param property_name: Name of memory property
        :param byte_count: Number of bytes to read.
        :param start_byte_index: Memory byte index, to start reading data. Must be on word boundary.
        :param word_byte_length: Number of bytes per memory word.
        :return: bytearray holding read data.
        """
        service, done_cb = self.make_done(done)
        token = service.refresh_memory(
            self.ctx, property_name, byte_count, start_byte_index, word_byte_length, done_cb
        )
        return self.add_pending(token)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
from collections import namedtuple, defaultdict, deque
from typing import List

from chipscopy.api.memory import Memory
from chipscopy.client.core import CoreParent
from chipscopy.client.debug_core_client import DebugCoreClientNode
from chipscopy.client.jtagdevice import JtagDevice
from chipscopy.client.mem import MemoryNode
from chipscopy.client.view_info import ViewInfo

# Tuple with the raw data for tracking scanned device information for a device
# Used during DeviceScanner scan
# TODO: Add an incrementing device number for easy path reporting (not jtag_index)
# Tuple with the raw data for tracking scanned device information for nodes associated with the device
# Used during DeviceScanner scan

_DeviceIdentificationTuple = namedtuple(
    "DeviceTracker",
    ["family", "node_identification", "dna", "cable_context", "cable_name", "jtag_index"],
)
_DeviceNodeIdentificationTuple = namedtuple(
    "NodeTracker", ["server_type", "view_name", "name", "context", "hier_name"]
)


class DeviceIdentification:
    """DeviceIdentification is a wrapper around the _DeviceIdentificationTuple object providing some additional
    services given the raw device identification data.

    All devices are expected to carry some form of identification. DNA is the most reliable, but we take what we can
    get for a likely match. Jtag cable and jtag position will work if DNA is unavailable for non-Versal devices.
    This identification helps us match up nodes targets across different views, especially ones that don't make
    hierarchical sense like DPC and Versal top level targets on the same device.

    TODO: Get rid of the _DeviceIdentificationTuple - just use this class for everything. That should be simpler.
    """

    def __init__(self, device_tracker: _DeviceIdentificationTuple):
        self.device_tracker = device_tracker

    def __getitem__(self, key):
        return getattr(self.device_tracker, key)

    def __repr__(self):
        # Make a nice nested dictionary representation of the nested named tuples.
        # This necessary because namedtuple._asdict() doesn't work well with nesting tuples.
        # I want the inner node_identification to be a list of dicts, not a list of tuples
        outer_dict = defaultdict(list)
        for field, value in zip(self.device_tracker._fields, self.device_tracker):
            if isinstance(value, list):
                for device_node_id_tuple in value:
                    # Nested list of tuples
                    inner_dict = {}
                    for field2, value2 in zip(device_node_id_tuple._fields, device_node_id_tuple):
                        inner_dict[field2] = value2
                    outer_dict[field].append(inner_dict)
            else:
                outer_dict[field] = value
        raw_json = json.dumps(outer_dict, indent=4)
        return raw_json

    def __str__(self):
        json_obj = self.__repr__()
        d = json.loads(json_obj)
        family = d.get("family", "unknown_family")
        dna = hex(d.get("dna", "0x0"))
        return f"{family}:{dna}"

    def find_top_level_device_node(self, hw_server):
        """
        Returns the device node we will use as a top level to identify
        this device. This node is where we also hang additional device services
        (wrappers). This forces the additional device services lifecycle
        to match the device node. Helps keep everything in sync.
        It really doesn't matter which device node is picked, as long as it
        is consistent everytime we call.
        """
        try:
            node = self.find_jtag_node(hw_server)
        except ValueError:
            # Jtag node finder didn't work - happens in case like XVC with
            # no jtag. Fall back to debugcore instead in that case.
            node = None

        if node is None:
            try:
                node = self.find_debugcore_node(hw_server)
            except ValueError:
                # Could not find the node - this is not good.
                node = None

        if node is None:
            raise ValueError("Failed to lookup top level device node")
        return node

    def get_target_tuples(self) -> List[_DeviceNodeIdentificationTuple]:
        """
        Returns: Tuple with all the accessible targets for this device including a hierarchical style path.
        """
        return self.device_tracker.node_identification

    def _find_node(self, view_name, server, node_class_type):
        """Finds the top level node for a view given the name and server.
        device_node_ctx_finder() return defaults to DPC. This is only
        for the chipscope and debugcore views. Use the find_jtag_node
        for jtag because the required process is slightly different.
        """
        assert view_name == "chipscope" or view_name == "debugcore"
        view = server.get_view(view_name)
        discovery_node_ctx = self.device_node_ctx_finder(view_name)
        discovery_node = view.get_node(discovery_node_ctx, node_class_type)
        return discovery_node

    def find_debugcore_node(self, server) -> DebugCoreClientNode:
        return self._find_node("debugcore", server, DebugCoreClientNode)

    def find_chipscope_node(self, server) -> CoreParent:
        return self._find_node("chipscope", server, CoreParent)

    def find_jtag_node(self, hw_server) -> JtagDevice:
        """Finds the node we use for programming the device.
        Returns: jtag programming node for the device
        """
        # Go through list of possible jtag configurable devices - when we find our matching context,
        # program it. This is versal specific. It will need work when we add ultrascale.
        # One problem at a time...
        jtag_view = hw_server.get_view("jtag")
        jtag_node = None
        # Grab the jtag node context stored for this device jtag view.
        discovery_node_ctx = self.device_node_ctx_finder("jtag", ".*")
        for node in jtag_view.get_all():
            if JtagDevice.is_compatible(node):
                jtag_node = jtag_view.get_node(node.ctx, JtagDevice)
                if jtag_node.ctx == discovery_node_ctx:
                    break
        return jtag_node

    def find_dpc_memory_node(self, hw_server):
        memory_view: ViewInfo = hw_server.get_view("memory")
        top_memory_node_ctx = self.device_node_ctx_finder("memory", "^DPC")
        node = memory_view.get_node(top_memory_node_ctx, MemoryNode)
        upgraded_node = Memory.check_and_upgrade(memory_view, node)
        return upgraded_node

    def find_versal_node(self, hw_server):
        memory_view: ViewInfo = hw_server.get_view("memory")
        top_memory_node_ctx = self.device_node_ctx_finder("memory", "^Versal")
        node = memory_view.get_node(top_memory_node_ctx, MemoryNode)
        return node

    def get_memory_target_nodes(self, hw_server):
        # Return the memory target nodes for this device
        view_name = "memory"
        node_ctx_list = list()
        access_priority_list = [r"^DPC", r"^Versal", r"^XVC"]
        for access_type in access_priority_list:
            for node_tracker in self.device_tracker.node_identification:
                if (
                    re.search(access_type, node_tracker.name)
                    and node_tracker.view_name == view_name
                ):
                    # The top level debug core discovery node
                    node_ctx_list.append(node_tracker.context)

        memory_view: ViewInfo = hw_server.get_view(view_name)
        memory_node_list = list()
        for node_ctx in node_ctx_list:
            node = memory_view.get_node(node_ctx)
            if MemoryNode.is_compatible(node):
                memory_node = memory_view.get_node(node.ctx, MemoryNode)
                if memory_node and len(memory_node.props.get("Name")) > 0:
                    memory_node_list.append(memory_node)

        # Add child nodes
        nodes_to_travel = deque(memory_node_list)
        memory_node_list = list()
        while nodes_to_travel:
            node = nodes_to_travel.popleft()
            if MemoryNode.is_compatible(node):
                memory_node = memory_view.get_node(node.ctx, MemoryNode)
                memory_node_list.append(memory_node)
                nodes_to_travel.extendleft(memory_view.get_children(memory_node))
        # Upgrade all
        upgraded_memory_node_list = [
            Memory.check_and_upgrade(memory_view, node) for node in memory_node_list
        ]
        return upgraded_memory_node_list

    def find_memory_target(self, hw_server, target: str):
        if target is None:
            dpc_or_versal_target = None
            target = "DPC"
        elif target == "DPC":
            # Use the DPC (Default if None is specified)
            dpc_or_versal_target = "^DPC"
        else:
            # Use the DAP
            dpc_or_versal_target = "^Versal"

        memory_view: ViewInfo = hw_server.get_view("memory")
        top_memory_node_ctx = self.device_node_ctx_finder("memory", dpc_or_versal_target)
        top_memory_node = memory_view.get_node(top_memory_node_ctx, MemoryNode)
        assert top_memory_node is not None

        # Now we have a top memory node for this device - either Versal or DPC. Do a search through all children
        # and stop when we find a matching target.
        target_node = None
        nodes_to_travel = deque([top_memory_node])
        while nodes_to_travel:
            node = nodes_to_travel.popleft()
            if MemoryNode.is_compatible(node):
                memory_node = memory_view.get_node(node.ctx, MemoryNode)
                if memory_node:
                    name = memory_node.props.get("Name")
                    if name and re.search(target, name):
                        target_node = memory_node
                        break
                    nodes_to_travel.extendleft(memory_view.get_children(memory_node))
        if not target_node:
            raise KeyError(f"Could not find target key {target} in memory targets")
        target_node = Memory.check_and_upgrade(memory_view, target_node)
        return target_node

    def device_node_ctx_finder(self, view_name: str, target_name_regex: str = None):
        """This is a function that returns the correct top level CoreParent context used for debug core scanning.
        It should return the correct context for a device given the view and target name.
        The returned node will be used when calling setup_debug_cores and device programming.

        Args:
            view_name: view to search (jtag, memory, debugcore, chipscope)
            target_name_regex: regex to override default target prioritized search

        Returns: top level node for this device (or ValueError exception if none found)
        """
        if target_name_regex:
            access_priority_list = [target_name_regex]
        else:
            access_priority_list = [r"^DPC", r"^Versal", r"^XVC", r"^Ultra"]
        for access_type in access_priority_list:
            for node_tracker in self.device_tracker.node_identification:
                if (
                    re.search(access_type, node_tracker.name)
                    and node_tracker.view_name == view_name
                ):
                    # The top level debug core discovery node
                    return node_tracker.context
        # Should not happen.
        raise (
            ValueError(
                f"Could not find matching node - view:{view_name}, target:{target_name_regex}"
            )
        )



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import math
import re
from math import exp, log
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

from chipscopy.utils.printer import printer

try:
    import kaleido
    import plotly as px
    import plotly.graph_objs as go

    _plotly_available = True
except ImportError:
    _plotly_available = False

try:
    from IPython import get_ipython
    from IPython.display import Image, display

    _jupyter_available = True
except ImportError:
    get_ipython = None
    Image = None
    display = None
    _jupyter_available = False

from chipscopy.api.ibert.aliases import EYE_SCAN_HORZ_RANGE

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.api.ibert.eye_scan import EyeScan


def check_for_plotly():
    if not _plotly_available:
        raise ImportError(
            f"Plotting packages not installed! Please run 'pip install chipscopy[core-addons]'"
        )


def check_for_jupyter():
    if not _jupyter_available:
        raise ImportError(
            f"Jupyter packages not installed! Please run 'pip install chipscopy[jupyter]'"
        )


def is_running_notebook():
    try:
        check_for_jupyter()
        shell = get_ipython().__class__.__name__
    except ImportError:
        shell = None
    except NameError:
        shell = None
    if shell == "ZMQInteractiveShell":
        return True
    else:
        return False


class EyeScanPlot:
    """
    Container for eye scan plot
    """

    def __init__(self, eye_scan):
        self.eye_scan: EyeScan = eye_scan
        """Link to the `EyeScan` object"""

        self.fig: go.Figure = None
        """Plotly `Figure` object"""

        self._x: List[float] = list()
        self._y: List[int] = list()
        self._z: List[List[float]] = list()

    def _clear_out_data(self):
        self._x, self._y, self._z = list(), list(), list()

    def _generate(self, *, title=""):
        if title == "":
            title = f"{self.eye_scan.rx.handle} ({self.eye_scan.name})"

        self._clear_out_data()

        if self.eye_scan.scan_data.processed is None:
            raise RuntimeError(
                f"No plot data received from server! "
                f"This might be because the scan did not finish successfully."
            )

        xs, ys = set(), set()
        for point in self.eye_scan.scan_data.processed.scan_points:
            xs.add(point[0])
            ys.add(point[1])

        self._x = sorted(xs)
        self._y = sorted(ys)

        # This will be needed later on to generate the legend
        min_ber, max_ber, plot_z_to_ber = -1234, -1234, list()

        for y in self._y:
            row_vals = list()
            z_hovertext_row = list()
            for x in self._x:
                ber = self.eye_scan.scan_data.processed.scan_points[(x, y)].ber
                val = log(ber, 10)
                hovertext = format(pow(10, val), ".2e")
                z_hovertext_row.append(hovertext)
                row_vals.append(val)

            curr_min = min(row_vals)
            if min_ber == -1234 or curr_min < min_ber:
                min_ber = curr_min

            curr_max = max(row_vals)
            if max_ber == -1234 or curr_max > max_ber:
                max_ber = curr_max

            self._z.append(row_vals)
            plot_z_to_ber.append(z_hovertext_row)

        extracted_data = re.match(
            r"^(.*) UI to (.*) UI$", self.eye_scan.scan_data.all_params[EYE_SCAN_HORZ_RANGE]
        )
        max_horz_range_ui, min_horz_range_ui = (
            float(extracted_data.group(1)),
            float(extracted_data.group(2)),
        )

        max_horz_range_codes, min_horz_range_codes = max(self._x), min(self._y)
        for index, x_in_codes in enumerate(self._x):
            self._x[index] = min_horz_range_ui + (
                ((x_in_codes - min_horz_range_codes) * (max_horz_range_ui - min_horz_range_ui))
                / (max_horz_range_codes - min_horz_range_codes)
            )

        # Generate BER to z mapping for colorbar a.k.a legend
        colorbar_number_of_ticks: int = math.floor(min_ber)

        incr_factor = (max_ber - min_ber) / (colorbar_number_of_ticks - 1)

        color_bar_tick_text = list()
        color_bar_tick_values = list()
        # for x_factor in range(colorbar_number_of_ticks):
        #     val = min_ber + (x_factor * incr_factor)
        #     color_bar_tick_values.append(val)
        #     tick_text = format(pow(base, val) if base != 0 else exp(val), ".2e")
        #     color_bar_tick_text.append(tick_text)
        for i in range(-1, math.floor(min_ber) - 1, -1):
            color_bar_tick_text.append(format(pow(10, i), ".0e"))
            color_bar_tick_values.append(i)

        contour = go.Contour(
            x=self._x,
            y=self._y,
            z=self._z,
            line=dict(smoothing=0.85, width=0),
            colorbar=dict(
                nticks=abs(math.floor(min_ber)),
                ticks="outside",
                tickmode="array",
                tickvals=color_bar_tick_values,
                ticktext=color_bar_tick_text,
                title=dict(text="BER"),
            ),
            contours=dict(start=-1, end=math.floor(min_ber), size=0.5),
            colorscale="portland",
            hovertext=plot_z_to_ber,
            hoverinfo="x+y+text",
        )

        x_axis_options = dict(zeroline=False, title="UI", mirror=True)
        y_axis_options = dict(zeroline=False, title="Voltage (Codes)", mirror=True)
        layout = go.Layout(xaxis=x_axis_options, yaxis=y_axis_options)

        self.fig = go.Figure(layout=layout, data=contour)

        self.fig.update_layout(title=dict(text=title))

    def show(self, display_type: str = "automatic", *, title: str = ""):
        """
        Displays the plot in the browser

        Args:
            display_type: image format to show. Options are automatic, static, dynamic. Default is 'automatic'
            title: title to display in plot

        """
        check_for_plotly()

        self._generate(title=title)

        if display_type == "automatic":
            if is_running_notebook():
                display_type = "static"
            else:
                display_type = "dynamic"

        if display_type == "dynamic":
            self.fig.show()
        elif display_type == "static":
            check_for_jupyter()
            image_bytes = self.fig.to_image(format="png")
            ipython_image = Image(image_bytes)
            display(ipython_image)
        else:
            raise ValueError("show: display_type must be automatic, static, or dynamic")

    def save(
        self, path: str = ".", file_name: str = None, *, file_format: str = "svg"
    ) -> Optional[str]:
        """
        Save plot to a file

        Args:
            path: **(Optional)** Location where file should be saved.

            file_name: **(Optional)** Name of the file

            file_format: **(Optional)** File format. Default is `SVG`

        Returns:
            Path of the saved plot

        """
        check_for_plotly()

        output_path = Path(path)

        if file_name is None:
            file_name = f"{self.eye_scan.name}"

        output_path = output_path.joinpath(f"{file_name}.{file_format}")

        self._generate()
        self.fig.write_image(str(output_path.resolve()), height=1080, width=1920, scale=1)

        printer(f"Saved eye scan plot to {str(output_path.resolve())}", level="info")

        return str(output_path.resolve())


class EyeScanImporter:  # pragma: no cover
    @staticmethod
    def load_from_vivado_csv(file_path: Path = None):
        if file_path is None:
            file_path = Path(
                "C:/Users/araju/Documents/Pycharm Projects/chipscopy/chipscopy/api/"
                "ibert/scan/plots/vivado_eye_scan.csv"
            )

        def normalize_codes_to_ui(data, min_range, max_range):
            min_element, max_element = min(data), max(data)
            for index, value in enumerate(data):
                data[index] = min_range + (
                    ((value - min_element) * (max_range - min_range)) / (max_element - min_element)
                )

        x_axis, y_axis, scan_data = list(), list(), list()
        with file_path.open("r") as csv_file:
            csv_file_contents = csv.reader(csv_file)
            for row in csv_file_contents:
                if row[0] == "Horizontal Range":
                    split = row[1].split()
                    min_value_range, max_value_range = (float(split[0]), float(split[-2]))
                elif row[0] == "Scan Start":
                    for index, row in enumerate(csv_file_contents):
                        if row[0] == "Scan End":
                            break
                        if not x_axis:
                            x_axis = [float(value) for value in row[1:]]
                            normalize_codes_to_ui(x_axis, min_value_range, max_value_range)
                            continue
                        y_axis.append(float(row[0]))
                        scan_data.append([log(float(value)) for value in row[1:]])

        return x_axis, y_axis, scan_data



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

"""
.. |state| replace::   :class:`MILAState`
.. |tc|    replace::   :class:`MILATriggerCondition`
.. |tpos|  replace::   :class:`MILATriggerPosition`
.. |csel|  replace::   :class:`MILAClockSelect`
.. |dsel|  replace::   :class:`MILADataSelect`

MILA Service Overview
---------------------

The MILA Service communicates with DDRMC Mini ILA cores.

*Data Model*

The MILA model is kept in "properties". The properties are divided into groups:

* `Static Config Properties`_ - Core capabilities and dimensions.
* `Status Properties`_ - Current core dynamic state.
* `Control Properties`_ - Control of the core.
* `Data Properties`_ - Uploaded waveform.
* `Data Select Properties`_ - Input Data Probe Selection MUX and associated properties.
* `Mux Local Enable Properties`_ - Mux Node local input enable vs enable of previous Mux Node lane input.
* `Mux Local Select Properties`_ - Mux Node local input selection per lane, when lane has local input enabled.


*Methods*

The Service methods are of the following types:

* `Property Methods`_ - Set/get/reset/report properties.
* `Probe Methods`_ - Create/delete probes. Set/get probe match values.
* `Mux Tree Methods`_ - Setup/report the Mux Tree Node settings.
* `Core Communication Methods`_ - Read/write properties from/to core. Arm the core and upload waveform.

Service Name
------------
.. autodata:: NAME

Properties
----------
Property Groups
^^^^^^^^^^^^^^^
Properties are divided into groups. Properties within a group are often have
their values 'set' or 'get' at the same time.

+----------------------------------+-------+------------------------------------------+
| Name                             | Type  | Description                              |
+==================================+=======+==========================================+
| PROP_GROUP_STATIC_INFO           | |str| | Read-only properties giving              |
|                                  |       | capabilities and dimensions.             |
+----------------------------------+-------+------------------------------------------+
| PROP_GROUP_STATUS                | |str| | Read-only properties showing current     |
|                                  |       | dynamic state.                           |
+----------------------------------+-------+------------------------------------------+
| PROP_GROUP_CONTROL               | |str| | Write/Read properties to control core.   |
+----------------------------------+-------+------------------------------------------+
| PROP_GROUP_DATA                  | |str| | Read-only properties holding waveform.   |
+----------------------------------+-------+------------------------------------------+
| PROP_GROUP_DATA_SELECT           | |str| | Properties for register ila_cntrl.       |
+----------------------------------+-------+------------------------------------------+
| PROP_GROUP_MUX_LOCAL_ENABLE      | |int| | Mux select between local/lane inputs.    |
|                                  |       | One value {0,1} per lane.                |
+----------------------------------+-------+------------------------------------------+
| PROP_GROUP_MUX_LOCAL_SELECT      | |int| | Mux select which input signal "group",   |
|                                  |       | when MUX_LOCAL_ENABLE, selects "local".  |
|                                  |       | One value {0-7} per lane.                |
+----------------------------------+-------+------------------------------------------+

Static Config Properties
^^^^^^^^^^^^^^^^^^^^^^^^^^
*Static Config* are read-only properties giving the dimensions and feature set of
the core.

+--------------------------+--------+-----------------------------------------+
| Name                     | Type   | Description                             |
+==========================+========+=========================================+
| PROP_DATA_DEPTH          | |int|  | Sample depth of waveform data.          |
+--------------------------+--------+-----------------------------------------+
| PROP_DATA_WIDTH          | |int|  | Bit width of waveform data.             |
+--------------------------+--------+-----------------------------------------+
| PROP_HAS_CAPTURE_CONTROL | |bool| | True if core has capture control.       |
+--------------------------+--------+-----------------------------------------+
| PROP_HAS_CIN             | |bool| | True if core has carry in port.         |
+--------------------------+--------+-----------------------------------------+
| PROP_HAS_COUT            | |bool| | True if core has carry out port.        |
+--------------------------+--------+-----------------------------------------+
| PROP_HAS_MU              | |bool| | True if core has match units            |
+--------------------------+--------+-----------------------------------------+
| PROP_HAS_STORAGE         | |bool| | True if core stores waveform data.      |
+--------------------------+--------+-----------------------------------------+
| PROP_HAS_SYNC            | |bool| | True if synchronized logic is included  |
|                          |        | to control control input ports.         |
+--------------------------+--------+-----------------------------------------+
| PROP_HAS_TRIGGER_IN      | |bool| | True if core has trigger in port.       |
+--------------------------+--------+-----------------------------------------+
| PROP_HAS_TRIGGER_OUT     | |bool| | True if core has trigger out port.      |
+--------------------------+--------+-----------------------------------------+
| PROP_TRIGGER_LEVELS      | |int|  | Number of sequential trigger levels.    |
+--------------------------+--------+-----------------------------------------+

Status Properties
^^^^^^^^^^^^^^^^^^^
*Status* properties are read-only, which show the dynamic atatus of the core.

+--------------------------+--------+-----------------------------------------+
| Name                     | Type   | Description                             |
+==========================+========+=========================================+
| PROP_CARRY               | |bool| | True if all match units match.          |
+--------------------------+--------+-----------------------------------------+
| PROP_DONE                | |bool| | True if waveform data buffer is full.   |
+--------------------------+--------+-----------------------------------------+
| PROP_SAMPLE_COUNT        | |int|  | Current number of captured samples.     |
+--------------------------+--------+-----------------------------------------+
| PROP_STATE               | |state|| State of capture controller.            |
+--------------------------+--------+-----------------------------------------+
| PROP_TRIGGER_SAMPLE      | |int|  | Trigger location within capture buffer. |
|                          |        | Used by service to organize waveform    |
|                          |        | data.                                   |
+--------------------------+--------+-----------------------------------------+

Control Properties
^^^^^^^^^^^^^^^^^^^^
*Control* read/write properties are used to control the core.
Some control properties have list value, with one entry per match unit.
The ILA has 3 match units, which can work in sequence A -> B -> C.

+--------------------------+--------------+-----------------------------------------+
| Name                     | Type         | Description                             |
+==========================+==============+=========================================+
| PROP_ARM                 | |bool|       | Set to True, to arm the core.           |
+--------------------------+--------------+-----------------------------------------+
| PROP_CARRY_IN_ENABLE     | List[|str|]  | Enable carry-out hold value.            |
|                          |              | One char string per match unit [A,B,C]  |
|                          |              | see :ref:`prop_match_value_mm_table`    |
+--------------------------+--------------+-----------------------------------------+
| PROP_CARRY_OUT_HOLD      | List[|bool|] | Enable carry-out hold value.            |
|                          |              | One value per match unit [A,B,C]        |
+--------------------------+--------------+-----------------------------------------+
| PROP_CLK_SEL             | |csel|       | Set clock sel mux for probe input       |
+--------------------------+--------------+-----------------------------------------+
| PROP_ENABLE              | |bool|       | Set to False, to reset core.            |
+--------------------------+--------------+-----------------------------------------+
| PROP_MATCH_VALUE         | List[|str|]  | Bit values in a string. MSB to LSB.     |
|                          |              | For valid bit values:                   |
|                          |              | see :ref:`prop_match_value_mm_table`    |
|                          |              | One string value per match unit [A,B,C] |
+--------------------------+--------------+-----------------------------------------+
| PROP_TRIGGER_POSITION    | |tpos|       | Trigger position within trace buffer.   |
+--------------------------+--------------+-----------------------------------------+
| PROP_WINDOW_SIZE         | |int|        | Sample count per window. The value must |
|                          |              | be a power-of-two. Property value only  |
|                          |              | applied when trigger position is a zero.|
+--------------------------+--------------+-----------------------------------------+
| PROP_TRIGGER_CONDITION   | List[|tc|]   | Trigger condition: AND, OR.             |
|                          |              | One value per match unit [A,B,C]        |
+--------------------------+--------------+-----------------------------------------+

Data Properties
^^^^^^^^^^^^^^^^^
The *Data* read-only properties represent the uploaded waveform.

The samples in the trace_data are aligned on byte boundaries.

This formula can be used to read a bit from the trace_data:
::

    bytes_per_sample = len(trace_data) // trace_sample_count

    def get_bit_value(trace_data: bytearray, bytes_per_sample: int, sample_index: int, data_bit_index: int) -> bool:
        byte_value = trace_data[sample_index * bytes_per_sample + data_bit_index // 8]
        mask = 1 << (data_bit_index & 0x7)
        return (byte_value & mask) != 0

+--------------------------+-----------------+--------------------------------+
| Name                     | Type            | Description                    |
+==========================+=================+================================+
| trace_data               | |bytearray|     | Waveform bit values.           |
+--------------------------+-----------------+--------------------------------+
| trace_width              | |int|           | Data sample bit width.         |
+--------------------------+-----------------+--------------------------------+
| trace_sample_count       | |int|           | Sample count.                  |
+--------------------------+-----------------+--------------------------------+
| trace_trigger_position   | List[|int|]     | List of trigger positions.     |
|                          |                 | One value, per window.         |
+--------------------------+-----------------+--------------------------------+
| trace_window_size        | |int|           | Sample count per window.       |
+--------------------------+-----------------+--------------------------------+

Data Select Properties
^^^^^^^^^^^^^^^^^^^^^^
The *Data Select* properties represent the device ila_capture_data register.
These properties are read from the device when the MILA Service is initialized and written to the device
during the 'arm' command. Property data_sel, is set by method mux_select_inputs.

+----------------+--------+---------------------------------------------------------------------+
| Name           | Type   | Description                                                         |
+================+========+=====================================================================+
| sel_fifo_empty | |int|  | XPI related.                                                        |
+----------------+--------+---------------------------------------------------------------------+
| phy_nib_sel    | |int|  | Physical nibble selection for GT Status or signals mapped physically|
+----------------+--------+---------------------------------------------------------------------+
| channel_sel    | |int|  | Channel selection when used for XPI selections of debug data.       |
+----------------+--------+---------------------------------------------------------------------+
| byte_sel       | |int|  | Byte selection when used for Cal/XPI selections of debug dataC      |
+----------------+--------+---------------------------------------------------------------------+
| data_sel       | |dsel| | ila_capture_data_sel mux selection.                                 |
+----------------+--------+---------------------------------------------------------------------+

Mux Local Enable Properties
^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. line-block::

    The *Mux Local Enable* properties choose for each Mux Node output lane, between
    local input or input from previous Mux Node.
    Default value is [0,0,0,0,0,0,0,0]
    which means that none of the 8 local input data chunks are selected.
    With the default value, all Mux Node lanes are feedthroughs from previous Mux Node.

+----------------+-------------+-------------------------------------------+
| Name           | Type        | Description                               |
+================+=============+===========================================+
|\*.local_enable | List[|int|] | 8 values. '1' for local input.            |
|                |             | '0' choose previous Mux Node lane output. |
+----------------+-------------+-------------------------------------------+


Mux Local Select Properties
^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. line-block::

    The *Mux Local Select* properties select which local inputs to connect to lane outputs,
    when a lane local enable is '1'.
    Default value for these properties are [0,0,0,0,0,0,0,0].

+----------------+--------------+---------------------------------------------------------------------+
| Name           | Type         | Description                                                         |
+================+==============+=====================================================================+
|\*.local_select | List[|int|]  | 8 values. When local enabled for output lane, selects local inputs. |
+----------------+--------------+---------------------------------------------------------------------+



Property Value Types
--------------------

.. _prop_match_value_mm_table:

Match Value Bits
^^^^^^^^^^^^^^^^
=====  ===========================
Value  Description
=====  ===========================
'0'    Zero.
'1'    One.
'R'    Rising edge. Transition '0' to '1'.
'F'    Falling edge. Transition '1' to '0'.
'B'    Both. Either rising or falling.
'N'    No edge. No transition.
'X'    Don't care. Any value.
'_'    Separator for readability.
=====  ===========================

_PropertyPermission Enum Values
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: chipscopy.client.core_property_client.PropertyPermission
    :members:
    :undoc-members:

Capture State Enum Values
^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: chipscopy.tcf.services.mila.MILAState
    :members:
    :undoc-members:

Trigger Position Enum Values
^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: chipscopy.tcf.services.mila.MILATriggerPosition
    :members:
    :undoc-members:

Trigger Condition Enum Values
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: chipscopy.tcf.services.mila.MILATriggerCondition
    :members:
    :undoc-members:

Param Types
-----------

PropertyValues
^^^^^^^^^^^^^^
.. autodata:: PropertyValues

Dict to hold property values.

::

    PropertyValues = Dict[str, Any]

=============== ================
Key             Value
=============== ================
property name   property value
=============== ================


.. _MILAProbeDefs:

MILAProbeDefs
^^^^^^^^^^^^^
.. autodata:: MILAProbeDefs

Dict type to hold probe definitions.
::

    MILAProbeDefs = Dict[str, List[int]]

============ ===============================================
Key          Value
============ ===============================================
probe name   List with left bit index, right bit index.`
============ ===============================================

Example: Two probes named *sensor* and *enable* which are 8 bits and 1 bit wide.
::

    pp = {
        'sensor': [15, 8],
        'enable': [2, 2]
    }



General Methods
---------------

getName
^^^^^^^
.. automethod:: MILAService.getName

connect
^^^^^^^
.. automethod:: MILAService.connect

initialize
^^^^^^^^^^
.. automethod:: MILAService.initialize

terminate
^^^^^^^^^
.. automethod:: MILAService.terminate

Property Methods
----------------

get_property
^^^^^^^^^^^^
.. automethod:: MILAService.get_property

get_property_group
^^^^^^^^^^^^^^^^^^
.. automethod:: MILAService.get_property_group

report_property
^^^^^^^^^^^^^^^
.. automethod:: MILAService.report_property

report_property_group
^^^^^^^^^^^^^^^^^^^^^
.. automethod:: MILAService.report_property_group

reset_property
^^^^^^^^^^^^^^
.. automethod:: MILAService.reset_property

reset_property_group
^^^^^^^^^^^^^^^^^^^^
.. automethod:: MILAService.reset_property_group

set_property
^^^^^^^^^^^^
.. automethod:: MILAService.set_property

set_property_group
^^^^^^^^^^^^^^^^^^
.. automethod:: MILAService.set_property_group


Probe Methods
-------------

define_probe
^^^^^^^^^^^^
.. automethod:: MILAService.define_probe

undefine_probe
^^^^^^^^^^^^^^
.. automethod:: MILAService.undefine_probe

get_probe
^^^^^^^^^
.. automethod:: MILAService.get_probe

get_probe_match_value
^^^^^^^^^^^^^^^^^^^^^
.. automethod:: MILAService.get_probe_match_value

set_probe_match_value
^^^^^^^^^^^^^^^^^^^^^
.. automethod:: MILAService.set_probe_match_value

Core Communication Methods
--------------------------

arm
^^^
.. automethod:: MILAService.arm

commit_property_group
^^^^^^^^^^^^^^^^^^^^^
.. automethod:: MILAService.commit_property_group

refresh_property_group
^^^^^^^^^^^^^^^^^^^^^^
.. automethod:: MILAService.refresh_property_group


upload
^^^^^^
.. automethod:: MILAService.upload


Mux Tree Methods
----------------

mux_add_node
^^^^^^^^^^^^
.. automethod:: MILAService.mux_add_node

mux_build_tree
^^^^^^^^^^^^^^
.. automethod:: MILAService.mux_build_tree

mux_select_inputs
^^^^^^^^^^^^^^^^^
.. automethod:: MILAService.mux_select_inputs

mux_get_selected_inputs
^^^^^^^^^^^^^^^^^^^^^^^
.. automethod:: MILAService.mux_get_selected_inputs

mux_commit
^^^^^^^^^^
.. automethod:: MILAService.mux_commit

mux_refresh
^^^^^^^^^^^
.. automethod:: MILAService.mux_refresh

mux_report
^^^^^^^^^^
.. automethod:: MILAService.mux_report


"""

import enum
from typing import Any, Dict, List, Union, Tuple
from chipscopy.tcf.services import Service, DoneHWCommand
from chipscopy.client.core_property_client import PropertyValues

# Property Group Names
PROP_GROUP_STATIC_INFO      = 'static_info'
PROP_GROUP_STATUS           = 'status'
PROP_GROUP_CONTROL          = 'control'
PROP_GROUP_DATA             = 'data'
PROP_GROUP_DATA_SELECT      = 'data_select'
PROP_GROUP_MUX_LOCAL_ENABLE = 'mux_local_enable'
PROP_GROUP_MUX_LOCAL_SELECT = 'mux_local_select'

# Static Info Property Names
PROP_DATA_DEPTH      = 'data_depth'
PROP_DATA_WIDTH      = 'data_width'
PROP_HAS_CAPTURE     = 'has_capture_control'
PROP_HAS_CIN         = 'has_cin'
PROP_HAS_COUT        = 'has_cout'
PROP_HAS_MU          = 'has_mu'
PROP_HAS_STORAGE     = 'has_storage'
PROP_HAS_SYNC        = 'has_sync'
PROP_HAS_TRIGGER_IN  = 'has_trigger_in'
PROP_HAS_TRIGGER_OUT = 'has_trigger_out'
PROP_LEVELS          = 'levels'

# Status Property Names
PROP_CARRY           = 'carry'
PROP_DONE            = 'done'
PROP_SAMPLE_COUNT    = 'sample_count'
PROP_STATE           = 'state'
PROP_TRIGGER_SAMPLE  = 'trigger_sample'

# Control Property Names
PROP_ARM                = 'arm'
PROP_CARRY_IN_ENABLE    = 'carry_in_enable'
PROP_CARRY_OUT_HOLD     = 'carry_out_hold'
PROP_CLK_SEL            = 'clk_sel'
PROP_ENABLE             = 'enable'
PROP_MATCH_VALUE        = 'match_value'
PROP_TRIGGER_POSITION   = 'trigger_position'
PROP_WINDOW_SIZE        = 'window_size'
PROP_TRIGGER_CONDITION  = 'trigger_condition'

# Data Property Names
PROP_TRACE_DATA             = 'trace_data'
PROP_TRACE_WIDTH            = 'trace_width'
PROP_TRACE_SAMPLE_COUNT     = 'trace_sample_count'
PROP_TRACE_TRIGGER_POSITION = 'trace_trigger_position'
PROP_TRACE_WINDOW_SIZE      = 'trace_window_size'

MILAProbeDefs = Dict[str, List[int]]
""" 
Dictionary type, to hold probe definitions.

Key: probe name
Value: Two item list with left_bit_index and right_bit_index.
"""


class MILAState(enum.Enum):
    """
    *Trigger states:*

    IDLE -> START -> PRE_TRIGGER -> TRIGGER -> POST_TRIGGER -> IDLE

    START is "starting capture (mostly here for one cycle)".
    PRE_TRIGGER is "Circular buffer is filling to the point where trigger is set"
    TRIGGER is "circular buffer is full and capture logic is looking for matching condition".
    When passing as values to/from API functions, use string values.
    """

    START = 0
    PRE_TRIGGER = 1
    TRIGGER = 2
    POST_TRIGGER = 3
    IDLE = 4


class MILATriggerPosition(enum.Enum):
    """
    The value of enum indicates percentage of buffer to capture before the trigger.
    Enum *EVERY_SAMPLE*, means every sample is a trigger sample.

    When passing as values to/from API functions, use string values.
    """

    FIRST = 0
    ONE_QUARTER = 25
    HALF = 50
    THREE_QUARTERS = 75
    LAST = 100
    EVERY_SAMPLE = 1000


class MILATriggerCondition(enum.Enum):
    """
    Trigger condition can be *AND* or *OR*.

    When passing as values to/from API functions, use string values.
    """

    OR = 0
    AND = 1


class MILAClockSelect(enum.Enum):
    """
    Selection of clock domain for ILA input data.

    When passing as values to/from API functions, use string values.
    """

    MC = 0
    INOC = 1


class MILADataSelect(enum.Enum):
    """
    Selection for Ila_capture_data_sel MUX.

    When passing as values to/from API functions, use string values.
    """

    DDRMC = 0
    XPI_WRITE = 1
    XPI_READ = 2
    CAL = 3
    RPI = 4
    FIVE = 5
    SIX = 6
    SEVEN = 7
    EIGHT = 8
    NINE = 9
    TEN = 10
    ELEVEN = 11
    TWELVE = 12
    THIRTEEN = 13
    FOURTEEN = 14
    FIFTEEN = 15


# Service name.
NAME = "MILA"
"""
MILA service name.
"""


class MILAService(Service):
    """TCF MILA Service interface."""

    def getName(self) -> str:
        """
        Get service name of this service.

        :returns: Service name :const:`NAME`.
        """
        return NAME

    def connect(self, ctx: str, target: Dict[str, Any], done: DoneHWCommand):
        """
        Connect to hw_server. This function does not belong here. Move to other "Service".

        :param ctx: Core context id.
        :param target: Values for 'HOST', 'PORT', 'CORE_OFFSETS'
        :param done: Callback with result and any error.
        """

    def initialize(self, ctx: str, done: DoneHWCommand) -> None:
        """
        Initialize data model for the MILA core.

        :param ctx: Core context id.
        :param done: Callback with result and any error.
        """

    def terminate(self, ctx: str, done: DoneHWCommand) -> None:
        """
        Remove data model for the MILA core.

        :param ctx: Core context id.
        :param done: Callback with result and any error.
        """

    def write32(self, ctx: str, write_addr: int, data: int, domain_index: int, done: DoneHWCommand) -> None:
        """
        Write 32 bit register.

        :param ctx: Core context id.
        :param write_addr: Address.
        :param data: 32 bit data
        :param domain_index: 0 for DDRMC_MAIN, 1 for DDRMC_NOC
        """

    def read32(self, ctx: str, read_addr: int, read_word_count: int, domain_index: int, done: DoneHWCommand) \
            -> Union[List[int], int]:
        """
        Read 32 bit register

        :param ctx: Core context id.
        :param read_addr: address to read from.
        :param read_word_count: number of 32 bit words.
        :param domain_index: 0 for DDRMC_MAIN, 1 for DDRMC_NOC
        :return int value if read_word_count == 1. Otherwise list of int values.
        """


    #
    # Property Methods
    #
    def get_property(self, ctx: str, property_names: List[str], done: DoneHWCommand)\
            -> PropertyValues:
        """
        Get values for properties.

        :param ctx: Core context id.
        :param property_names: List of property names. Empty list means get all properties.
        :param done: Callback with result and any error.
        :return: Property name/value pairs in a dict. See :data:`PropertyValues`.
        """
        raise NotImplementedError("Abstract method")

    def get_property_group(self, ctx: str, property_group_names: List[str], done: DoneHWCommand)\
            -> PropertyValues:
        """
        Get property values for property groups.

        :param ctx: Core context id.
        :param property_group_names: List of property group names. Empty list means all properties.
        :param done: Callback with result and any error.
        :return: Properties for named group names. :data:`PropertyValues`
        """
        raise NotImplementedError("Abstract method")

    def report_property(self, ctx: str, property_names: List[str], done: DoneHWCommand)\
            -> Dict[str, Dict[str, Any]]:

        """
        Return description of properties.

        Example: Description of property 'state'
        ::

            Attributes:

            group_name - Property group which the property belongs to.
            type - Property value type
            permission - Permissions control which methods can operate on property.
            name - Name of property
            default_value - Reset value for property.
            value - Current value of the property.

            {
                'state': {
                    'group_name': 'status',
                    'type': 'MILAState'),
                    'permission': 6,        # corresponds to bitwise enum value <_PropertyPermission.REFRESH|GET: 6>,
                    'name': 'state',
                    'default_value': 'IDLE' # corresponds to enum value <MILAState.IDLE: 4>,
                    'value': 'IDLE'         # <MILAState.IDLE: 4>)
                }
            }

        For further information about 'permission', see :class:`_PropertyPermission`

        :param ctx: Core context id.
        :param property_names: List of property names. Empty list means get all properties.
        :param done: Callback with result and any error.
        :return: Dict of property description dicts.
        """
        raise NotImplementedError("Abstract method")

    def report_property_group(self, ctx: str, groups: List[str], done: DoneHWCommand) \
            -> Dict[str, Dict[str, Any]]:
        """
        Return description of properties, which belong to argument 'groups'.

        :param ctx: Core context id.
        :param groups: List of property group names. Empty list means all properties.
        :param done: Callback with result and any error.
        :return: Dict of property description dicts.
        """
        raise NotImplementedError("Abstract method")

    def reset_property(self, ctx: str, property_names: List[str], done: DoneHWCommand) -> None:
        """
        Reset property values, to default values.

        :param ctx: Core context id.
        :param property_names: List of property names. Empty list means all resetable properties.
           Properties in property groups 'control' and 'data' are resettable,
           e.g. the trigger capture will be set to trigger on any value.
        :param done: Callback with result and any error.
        """

        raise NotImplementedError("Abstract method")

    def reset_property_group(self, ctx: str, groups: List[str], done: DoneHWCommand) -> None:
        """

        :param ctx: Core context id.
        :param groups:  List of property group names. Empty list means all resettable properties.
           Property groups 'control' and 'data' are resettable.
        :param done: Callback with result and any error.
        """
        raise NotImplementedError("Abstract method")

    def set_property(self, ctx: str, property_values: PropertyValues, done: DoneHWCommand) -> None:
        """
        Set properties. If no properties specified, do nothing.
        Validates properties values for correctness.

        :param ctx: Core context id.
        :param property_values: Property name/value pairs.
        :param done: Callback with result and any error.
        """
        raise NotImplementedError("Abstract method")

    def set_property_group(self, ctx: str, property_values: PropertyValues, done: DoneHWCommand) -> None:
        """
        Same behaviour as :meth:`set_property`

        :param ctx: Core context id.
        :param property_values: Property name/value pairs.
        :param done: Callback with result and any error.
        """

        self.set_property(ctx, property_values, done)

    #
    # Mux Tree Methods
    #
    def mux_add_node(self, ctx: str,
                     memory_domain: str,
                     addr: Union[int,None],
                     name: str,
                     long_name: str,
                     has_local_inputs: bool,
                     in0: str,
                     in1: str,
                     done: DoneHWCommand) -> None:
        """
        Defines a mux_node and declares which mux nodes feed into this mux_node.

        Example: Adding 3 mux nodes.
        ::

            # Using function alias for convenience.
            an = service.mux_add_node

            an(ctx, DDRMC_MAIN, 0x1450, 'ACAM0'  , 'd_0_13', in1='PT0'  )
            an(ctx, DDRMC_MAIN, 0x1454, 'DBUF0.1', 'd_0_14', in0='DBUF0.0', in1='ACAM0', has_local_inputs=False)
            an(ctx, DDRMC_MAIN, 0x1458, 'COM.0'  , 'd_0_15', in1='DBUF0.1')

        :param ctx: Core context id.
        :param memory_domain: DDRMC_MAIN or DDRMC_NOC
        :param addr: Address offset for mux node register
        :param name: Name of mux node. Must be unique.
        :param long_name: Description or other name.
        :param has_local_inputs: True if node has local inputs. False if "inputs" connected to other mux node.
        :param in0: Name of mux node, connecting to "data inputs", when no local inputs.
        :param in1: Name of mux node, connecting to channel inputs.
        """

    def mux_build_tree(self, ctx: str,
                       clk_sel_is_1_mux: str,
                       ila_capture_data_sel_muxes: [str],
                       done: DoneHWCommand) -> None:
        """
        .. line-block::

            Specifies which mux nodes, are connected to ila_cntrl.clk_sel and ila_capture_data_sel muxes.
            Builds tree data structure for the mux nodes.
            All calls to mux_add_node, need to be prior to calling this function.

            Defines local_enable, local_select properties for mux nodes which have register addresses.
            A mux node named 'xmpu' would get two registers: 'xmpu.in_mux' and 'xmpu.out_mux'.
            An in_mux property represents the local input selection mux.
            Default in_mux value is [False,False,False,False,False,False,False,False]
            An out_mux property represents the output channel selection mux.
            Default out_mux value is [0,1,2,3,4,5,6,7].
            The defaults correspond the mux node is a straight feed-through for the 8 channels.
            
        Example: Call to mux_build_tree
        ::

            service.mux_build_tree('ADEC',['COM.2', 'XPI_WRITE', 'XPI_READ', 'CAL', 'RPI'])


        :param ctx: Core context id.
        :param clk_sel_is_1_mux: Name of mux node, selected when ila_cntrl.clk_sel mux is '1'
        :param ila_capture_data_sel_muxes: list of mux nodes, connected to ila_capture_data_sel.
        """

    def mux_select_inputs(self, ctx: str, mux_node_inputs:[Tuple[str, int]],
                          done: DoneHWCommand) -> None:
        """
        Select in order which 16 bit input data chunks are connected to the MILA.
        The setting will be applied to the in_mux/out_mux properties.

        Example: Connect up first 4 input chunks from mux node 'MRS0' and every 2nd chunks from 'RETRY1'.
        ::

            service.mux_select_inputs([('MRS0', 0), ('MRS0', 1), ('MRS0', 2), ('MRS0', 3),
                                       ('RETRY1', 0), ('RETRY1', 2), ('RETRY1', 4), ('RETRY1', 6)])

        :param ctx: Core context id.
        :param mux_node_inputs: List of (node_name, input_chunk_index) pairs.
        """

    def mux_get_selected_inputs(self, ctx: str, done: DoneHWCommand) \
            -> [Union[None, Tuple[str, int]]]:
        """
        Based on property values, determined which input chunks are connected to the MILA.

        :param ctx: Core context id.
        :return: List of entry. Each entry is either a (node_name, input_chunk_index) pair or None for no input chunk.
        """

    def mux_commit(self, ctx: str, done: DoneHWCommand) -> None:
        """
        Write mux node '*.local_enable' and '*.local_select' property settings to device MILA registers.
        Note! ila_cntrl.clk_sel and ila_captur_data_sel settings are not written to MILA registers.

        :param ctx: Core context id.
        """

    def mux_refresh(self, ctx: str, done: DoneHWCommand) -> dict:
        """
        Read device mux node MILA registers to update mux node '*.local_enable' and '*.local_select' properties.
        Note! ila_cntrl.clk_sel and ila_capture_data_sel settings are not updated.

        :param ctx: Core context id.
        :return: local_enable/local_select property name-value pairs.
        """
    def mux_report(self, ctx: str, skip_default: bool, done: DoneHWCommand) -> str:
        """
        Report mux node settings, based on property values.

        :param ctx: Core context id.
        :param skip_default: If True, mux nodes set as "straight feed-throughs" are excluded from the report.
        :return: report as a text string.
        """


    #
    # Probe Methods
    #
    def define_probe(self, ctx: str, probe_defs: MILAProbeDefs, done: DoneHWCommand) -> None:
        """
        Create MILA probe(s), which are used to set/get a slice range
        of the MILA core property PROP_MATCH_VALUE.
        A probe has a left_index and a right_index, defining the bit range the probe use of the match word.

        Example: Two probes named *sensor* and *enable* which are 8 bits and 1 bit wide.
        ::

            pdefs = {
                'sensor': [15,8],
                'enable': [2,2]
            }
            service.create_probe( ctx, pdefs)

        :param ctx: Core context id.
        :param probe_defs: Property definitions.
        :param done: Callback with result and any error.
        """
        raise NotImplementedError("Abstract method")

    def undefine_probe(self, ctx: str, probe_names: List[str], done: DoneHWCommand) -> None:
        """
        Delete probes.

        :param ctx: Core context id.
        :param probe_names: List of probe names. Empty list means delete every probe.
        :param done: Callback with result and any error.
        """
        raise NotImplementedError("Abstract method")

    def get_probe(self, ctx: str, probe_names: List[str], done: DoneHWCommand) -> MILAProbeDefs:
        """
        Return probe definitions. Bit ranges of probes.

        :param ctx: Core context id.
        :param probe_names: Probe name or list of probe names. Empty list means every probe.
        :param done: Callback with result and any error.
        :return: See :ref:`MILAProbeDefs`. A Dict. Key is probe name. Value is 2-item list giving probe bit range.
        """
        raise NotImplementedError("Abstract method")

    def get_probe_match_value(self, ctx: str, probe_names: List[str], done: DoneHWCommand)\
            -> Dict[str, str]:
        """
        Get match match value for probe(s).

        :param ctx: Core context id.
        :param probe_names: Probe name list. Empty list means get all probe match values.
        :param done: Callback with result and any error.
        :return: Probe name/probe match value pairs.
        """
        raise NotImplementedError("Abstract method")

    def set_probe_match_value(self, ctx: str, match_pairs: Dict[str, str], done: DoneHWCommand) -> None:
        """

        :param ctx: Core context id.
        :param match_pairs: Probe name/probe match value pairs.
        :param done: Callback with result and any error.
        """
        raise NotImplementedError("Abstract method")

    #
    # Core Communication Methods
    #
    def arm(self, ctx: str, done: DoneHWCommand) -> None:
        """
        Arm the core with current *control* property settings.

        :param ctx: Core context id.
        :param done: Callback with result and any error.
        """
        raise NotImplementedError("Abstract method")

    def commit_property_group(self, ctx: str, groups: List[str], done: DoneHWCommand) -> None:
        """
        Write current property group values to the core.

        :param ctx: Core context id.
        :param groups: Group name or list of names.
            Empty list means property groups, which can be written to the core.
            Only the property group 'control' can be written to the MILA core.
        :param done: Callback with result and any error.
        """
        raise NotImplementedError("Abstract method")

    def refresh_property_group(self, ctx: str, groups: List[str], done: DoneHWCommand) -> PropertyValues:
        """
        Read values from the core, update MILA properties
        and return those property values to caller.

        :param ctx: Core context id.
        :param groups: List of group names.
            Empty list means all MILA refreshable groups.
            All MILA groups are refreshable, except 'static_info'.
        :param done: Callback with result and any error.
        :returns: Property name/value pairs.
        """
        raise NotImplementedError("Abstract method")

    def upload(self, ctx: str, done: DoneHWCommand) -> bool:
        """
        Checks if core trace buffer is full. If full will upload waveform and set property group 'Data'
        with waveform data.

        :param ctx: Core context id.
        :param done: Callback with result and any error.
        :return: True, if command uploaded a waveform.
        """
        raise NotImplementedError("Abstract method")



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from enum import Enum, unique
from pprint import pformat
from typing import Optional

from chipscopy import dm
from chipscopy.api._detail import dataclass_fields, filter_props


@unique
class CoreType(Enum):
    IBERT = "IBERT"
    NOC_ROOT = "NOC CORE"
    DDR = "DDRMC"
    AXIS_ILA = "AXIS ILA"
    AXIS_VIO = "AXIS VIO"
    AXIS_PCIE = "AXIS PCIE"
    AXIS_TRACE = "AXIS TRACE"
    AXI_DEBUG_HUB = "AXI_DEBUG HUB"
    SYSMON = "SYSMON"


@dataclass(frozen=True)
class CoreInfo:
    """Debug Fabric Core version information."""

    core_type: int
    """Core type number."""
    drv_ver: int
    """SW/IP interface version."""
    core_major_ver: int
    """Core IP major version."""
    core_minor_ver: int
    """Core IP minor version."""
    tool_major_ver: int
    """Major version of the design tool, which created the core."""
    tool_minor_ver: int
    """Minor version of the design tool, which created the core."""
    uuid: str
    """Core UUID. UUID is unique among debug cores in the design."""

    def __str__(self) -> str:
        return pformat(self.__dict__, 2)


CORE_INFO_MEMBERS = dataclass_fields(CoreInfo)


def get_core_info(tcf_node: dm.Node) -> Optional[CoreInfo]:
    if set(CORE_INFO_MEMBERS).issubset(tcf_node.props):
        # Only fabric cores have core info information.
        return CoreInfo(**filter_props(tcf_node.props, CORE_INFO_MEMBERS))

    return None



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from dataclasses import dataclass
from pprint import pformat
from typing import Dict, Any

from chipscopy.client.axis_ila_core_client import (
    ILATriggerCondition as TCF_ILATriggerCondition,
    ILATriggerMode as TCF_ILATriggerMode,
    ILACaptureCondition as TCF_ILACaptureCondition,
    ILACaptureMode as TCF_ILACaptureMode,
)
from chipscopy.api import dataclass_fields, filter_props
from chipscopy.shared.ila_util import (
    prop_to_enum,
    replace_enum_values,
)

""" Constant for indicating 'trigger in the middle of waveform window."""
ILA_TRIGGER_POSITION_HALF = -1

""" Constant for 'Maximum window sample depth."""
ILA_WINDOW_SIZE_MAX = -1


class ILACaptureCondition(enum.Enum):
    """
    For Basic Capture Control, the global boolean operator of participating probe comparators.

    ==================== ================================
    Enum Value           Description
    ==================== ================================
    ALWAYS               Capture all samples.
    AND                  AND probe comparators to filter data.
    OR                   OR probe comparators to filter data.
    NAND                 NAND probe comparators to filter data.
    NOR                  NOR probe comparators to filter data.
    ==================== ================================

    """

    ALWAYS = 4
    OR = 0
    AND = 1
    NOR = 2
    NAND = 3


class ILATriggerCondition(enum.Enum):
    """
    For Trigger Condition, set trigger mode including basic trigger global boolean operator
    of participating probe comparators

    =====================  ================================
    Enum Value             Description
    =====================  ================================
    AND                    Basic Trigger mode: AND probe comparators.
    OR                     Basic Trigger mode: OR probe comparators.
    NAND                   Basic Trigger mode: NAND probe comparators.
    NOR                    Basic Trigger mode: NOR probe comparators.
    IMMEDIATELY            Trigger now.
    TRIGGER_STATE_MACHINE  Advanced Trigger mode: TSM file decides when to trigger.
    =====================  ================================

    """

    OR = 0
    AND = 1
    NOR = 2
    NAND = 3
    IMMEDIATELY = 4
    TRIGGER_STATE_MACHINE = 5


class ILATrigInMode(enum.Enum):
    """
    =====================  ================================
    Enum Value             Description
    =====================  ================================
    DISABLED               No TRIG-IN port use.
    TRIG_IN_ONLY           Trigger only on TRIG-IN port.
    TRIGGER_OR_TRIG_IN     Trigger on basic/trigger mode or on TRIG-IN port.
    =====================  ================================

    """

    DISABLED = 0
    TRIG_IN_ONLY = 1
    TRIGGER_OR_TRIG_IN = 2


class ILATrigOutMode(enum.Enum):
    """
    Choices for Trig-Out port.

    ==================== ================================
    Enum Value           Description
    ==================== ================================
    DISABLED             Disables the TRIG_OUT port.
    TRIGGER_ONLY         Result of the basic/advanced trigger condition.
    TRIG_IN_ONLY         Propagates the TRIG_IN port to the TRIG_OUT port.
    TRIGGER_OR_TRIG_IN   OR-ing of the basic/advanced trigger and TRIG_IN port.
    ==================== ================================

    """

    DISABLED = 0
    TRIGGER_ONLY = 1
    TRIG_IN_ONLY = 2
    TRIGGER_OR_TRIG_IN = 3


class ILAState(enum.Enum):
    """
    ILA State Transitions:
        IDLE -> PRE_TRIGGER -> TRIGGER -> POST_TRIGGER -> IDLE

    =====================  ================================
    Enum Value             Description
    =====================  ================================
    IDLE                   Not armed.
    PRE_TRIGGER            Capturing pre-trigger samples.
    TRIGGER                Waiting for trigger.
    POST_TRIGGER           Capturing post-trigger samples.
    =====================  ================================

    """

    IDLE = 0
    PRE_TRIGGER = 1
    TRIGGER = 2
    POST_TRIGGER = 3


@dataclass(frozen=True)
class ILAControl:
    """Trigger setup"""

    capture_condition: ILACaptureCondition
    """Basic capture cntrol global boolean operator. See :class:`ILACaptureCondition`"""
    trig_in_mode: ILATrigInMode
    """Usage of TRIG-IN port. See :class:`ILATrigInMode`"""
    trig_out_mode: ILATrigOutMode
    """Usage of TRIG-OUT port. See :class:`ILATrigOutMode`"""
    trigger_condition: ILATriggerCondition
    """Global boolean operator, for trigger probes. See :class:`ILATriggerCondition`"""
    trigger_position: int
    """Trigger position index, within data window."""
    trigger_state_machine: str
    """Trigger State Machine, as a string."""
    window_count: int
    """Number of data windows."""
    window_size: int
    """Sample count for a window."""

    def __str__(self) -> str:
        return pformat(self.__dict__, 2)


ILA_CONTROL_MEMBERS = dataclass_fields(ILAControl)


@dataclass(frozen=True)
class ILAStatus:
    """Dynamic status of ILA capture controller"""

    capture_state: ILAState
    """Capture state. See :class:`ILAState`"""
    is_armed: bool
    """Trigger is armed."""
    is_full: bool
    """Data buffer is full."""
    is_trigger_at_startup: bool
    """ILA is armed at startup."""
    samples_captured: int
    """Number of samples captured in current data window."""
    windows_captured: int
    """Number of fully captured data windows."""
    samples_requested: int
    """Number of samples per window, as requested when was ILA armed."""
    windows_requested: int
    """Number of windows, as requested when ILA was armed."""
    trigger_position_requested: int
    """Trigger position, as requested when ILA was armed."""

    def __str__(self) -> str:
        return pformat(self.__dict__, 2)


ILA_STATUS_MEMBERS = dataclass_fields(ILAStatus)


# ILAStatus functions
def post_process_status(props: {}) -> None:
    props["capture_state"] = ILAState[props["capture_state"]]
    props["is_trigger_at_startup"] = props.get("is_tas", False)
    props["samples_requested"] = props.get("window_depth_readback")
    props["windows_requested"] = props.get("window_count_readback")
    props["trigger_position_requested"] = props.get("trigger_pos_readback")


def tcf_props_to_status(props: {}) -> ILAStatus:
    """ Make immutable ILAStatus instance from status dict."""
    post_process_status(props)
    return ILAStatus(**filter_props(props, ILA_STATUS_MEMBERS))


def tcf_refresh_status(tcf_node) -> ILAStatus:
    """Read dynamic status."""
    props = tcf_node.refresh_property_group(["status"])
    return tcf_props_to_status(props)


# ILAControl functions
__enum_control = (
    ("capture_condition", ILACaptureCondition),
    ("trig_out_mode", ILATrigOutMode),
    ("trigger_condition", ILATriggerCondition),
)


def control_from_tcf(props: {}) -> ILAControl:
    # No translation for trigger_position, window_count, window_size, trigger_out_mode
    cc = props.copy()

    is_tsm = props["trigger_mode"] in [
        TCF_ILATriggerMode.ADVANCED_OR_TRIG_IN.name,
        TCF_ILATriggerMode.ADVANCED_ONLY.name,
    ]
    is_trigger_or_trig_in = props["trigger_mode"] in [
        TCF_ILATriggerMode.BASIC_OR_TRIG_IN.name,
        TCF_ILATriggerMode.ADVANCED_OR_TRIG_IN.name,
        TCF_ILATriggerMode.TRIG_IN_ONLY.name,
    ]

    # trigger condition
    del cc["trigger_mode"]
    if is_tsm:
        cc["trigger_condition"] = ILATriggerCondition.TRIGGER_STATE_MACHINE

    # trig_in_mode
    if is_trigger_or_trig_in:
        cc["trig_in_mode"] = (
            ILATrigInMode.TRIGGER_OR_TRIG_IN if is_trigger_or_trig_in else ILATrigInMode.TRIG_IN
        )
    else:
        cc["trig_in_mode"] = ILATrigInMode.DISABLED

    # capture mode, capture condition
    del cc["capture_mode"]
    if props["capture_mode"] == TCF_ILACaptureMode.ALWAYS:
        cc["capture_condition"] = ILACaptureCondition.ALWAYS

    # trigger_state_machine not yet supported.
    cc["trigger_state_machine"] = ""

    prop_to_enum(cc, __enum_control)
    return ILAControl(**filter_props(cc, ILA_CONTROL_MEMBERS))


def control_to_tcf(cc: ILAControl) -> Dict[str, Any]:
    # trigger condition
    if cc.trigger_condition == ILATriggerCondition.TRIGGER_STATE_MACHINE:
        tcf_t_condition = TCF_ILATriggerCondition.AND.name
    else:
        tcf_t_condition = cc.trigger_condition.name

    # trigger mode
    if cc.trig_in_mode == ILATrigInMode.TRIG_IN_ONLY:
        tcf_t_mode = TCF_ILATriggerMode.TRIG_IN_ONLY.name
    elif cc.trigger_condition == ILATriggerCondition.TRIGGER_STATE_MACHINE:
        if cc.trig_in_mode == ILATrigInMode.TRIGGER_OR_TRIG_IN:
            tcf_t_mode = TCF_ILATriggerMode.ADVANCED_OR_TRIG_IN.name
        else:
            tcf_t_mode = TCF_ILATriggerMode.ADVANCED.name
    elif cc.trig_in_mode == ILATrigInMode.TRIGGER_OR_TRIG_IN:
        tcf_t_mode = TCF_ILATriggerMode.BASIC_OR_TRIG_IN.name
    else:
        tcf_t_mode = TCF_ILATriggerMode.BASIC_ONLY.name

    # capture mode, capture condition
    if cc.capture_condition == ILACaptureCondition.ALWAYS:
        tcf_capture_mode = TCF_ILACaptureMode.ALWAYS.name
        tcf_capture_condition = TCF_ILACaptureCondition.AND.name
    else:
        tcf_capture_mode = TCF_ILACaptureMode.BASIC
        tcf_capture_condition = cc.capture_condition.name

    res = {
        "capture_mode": tcf_capture_mode,
        "capture_condition": tcf_capture_condition,
        "trig_out_mode": cc.trig_out_mode.name,
        "trigger_mode": tcf_t_mode,
        "trigger_condition": tcf_t_condition,
        "trigger_position": cc.trigger_position,
        "window_count": cc.window_count,
        "window_size": cc.window_size,
    }
    replace_enum_values(res)

    return res



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import ClassVar, TYPE_CHECKING
from dataclasses import dataclass

from chipscopy.utils.printer import printer
from chipscopy.api import CoreType
from chipscopy.api._detail.debug_core import DebugCore

if TYPE_CHECKING:
    from chipscopy.client.axis_pcie_core_client import AxisPCIeCoreClient


@dataclass
class PCIe(DebugCore["AxisPCIeCoreClient"]):
    def __init__(self, pcie_tcf):
        super(PCIe, self).__init__(CoreType.AXIS_PCIE, pcie_tcf)

        self.name = pcie_tcf.props["Name"]
        self.uuid = self.core_info.uuid

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "uuid": self.uuid}

    def get_property(self, property_name_list=None):
        return_dict = {}
        items = self.get_properties()
        if (property_name_list is None) or (len(property_name_list) == 0):
            return_dict = items
        else:
            for key in property_name_list:
                if key in items:
                    return_dict[key] = items[key]
        return return_dict

    def get_properties(self):
        items = {"name": self.name, "uuid": self.uuid}
        items.update(self.core_tcf_node.props)
        return items

    def refresh(self):
        """
        Reads the PCIe debug memory again, and updated internal properties.

        Args:
            None

        Returns:
            None

        """
        self.core_tcf_node.read_data()
        self.get_property([])

    def read_data(self):
        """
        Reads the PCIe debug memory again, and updated internal properties, same as refresh()

        Args:
            None

        Returns:
            None

        """
        self.core_tcf_node.read_data()
        self.get_property([])

    def reset_core(self):
        """
        Resets the PCIe debug core, telling the IP to start collecting a new state trace

        Args:
            None

        Returns:
            None

        """
        self.core_tcf_node.reset_core()

    def __str__(self):
        return f"PCIe {self.name} ({self.uuid})"

    @staticmethod
    def _cap(s):
        return ".".join(i.capitalize() for i in s.split("."))

    def get_dot(self):
        """
        Returns a text string of the PCIe LTSSM in the DOT format. DOT is a graph
        description language:
        (https://en.wikipedia.org/wiki/DOT_(graph_description_language))
        This format can be graphed using python (networkx) or other graphing tools (graphviz
        and others)

        Args:
            None

        Returns:
            String with the PCIe LTSSM graph in a DOT format, using the same colors and labels
            as get_plt()

        """
        pos = {
            "Detect": 'pos="2,0!"',
            "Polling": 'pos="2,-2!"',
            "Configuration": 'pos="2,-4!"',
            "L0": 'pos="2,-6!"',
            "L1": 'pos="1,-7!"',
            "L0s": 'pos="2,-8!"',
            "R.Lock": 'pos="4,-8!"',
            "R.Speed": 'pos="4,-10!"',
            "R.Cfg": 'pos="6,-8!"',
            "R.Idle": 'pos="8,-8!"',
            "R.Eq": 'pos="6,-10!"',
            "Loopback": 'pos="4.5,-5.5!"',
            "Hotreset": 'pos="4.5,-4!"',
            "Disabled": 'pos="4.5,-2!"',
        }

        dot_src = "digraph {\n"
        dot_src += 'graph [pad="0.212,0.055" bgcolor=white splines=true esep="+40" sep="+50"]\n'
        dot_src += "node [style=filled]\n"
        d = self.get_property([])
        val_map = {}
        for state in pos.keys():
            s = "state." + state.replace(" ", "").lower()
            val_map[state] = "orange" if d[s] == 2 else "green" if d[s] == 1 else "gray"
            dot_src += f'"{state}" [fillcolor="{val_map[state]}" {pos[state]}]\n'

        for key in d.keys():
            if key.startswith("edge."):
                value = d[key]
                p = re.compile("edge.(.+)_(.+)")
                m = p.match(key)
                start = PCIe._cap(m.group(1))
                end = PCIe._cap(m.group(2))
                dot_src += f'  {start} -> {end} [label = "{value}"]\n'
        dot_src += "}"
        return dot_src

    def print_trace(self):
        """
        Prints PCIe trace to console, parsing the hierarchy of loops and substates to create
        a user-friendly output

        Args:
            None

        Returns:
            None

        """
        key, trace = self.property.get("state.trace").popitem()
        for t in trace:
            PCIe.print_bracketed(t, 0)

    @staticmethod
    def print_bracketed(trace, indent):
        colors = ["green", "blue", "magenta", "yellow", "blue"]
        color = colors[indent]
        if str(trace).isdigit():
            p = "\t" * indent + f"[{color}]{trace}[/{color}]"
            printer(p)
            return 0
        brace_locations = PCIe.find_brace(trace, "[]")
        start = str(trace).find("[")
        prev = 0
        end = -1
        while start != -1:

            end = brace_locations[start]
            items = trace[prev:start].split(",")
            for item in items:
                if item != "":
                    p = "\t" * indent + f"[{color}]{str(item).strip()}[/{color}]"
                    printer(p)
            PCIe.print_bracketed(trace[start + 1 : end], indent + 1)
            # skip past the ending bracket and comma
            prev = end + 1
            if prev >= len(trace):
                return
            if trace[prev] == ",":
                prev = prev + 1
            start = trace.find("[", prev + 1)
        else:
            items = str(trace[end + 1 :]).split(",")
            for item in items:
                if item != "":
                    p = "\t" * indent + f"[{color}]{str(item).strip()}[/{color}]"
                    printer(p)
        return 0

    @staticmethod
    def find_brace(s, braces):
        toret = {}
        pstack = []

        for i, c in enumerate(s):
            if c == braces[0]:
                pstack.append(i)
            elif c == braces[1]:
                if len(pstack) == 0:
                    raise IndexError("No matching closing parens at: " + str(i))
                toret[pstack.pop()] = i

        if len(pstack) > 0:
            raise IndexError("No matching opening parens at: " + str(pstack.pop()))

        return toret

    def get_plt(self):
        """
        Returns a matplotlib figure to plot, showing the PCIe LTSSM graph.  States will be
        colored green if they have been visited, orange if it's the last state visited, and
        grey if not visited.  The edge labels represent the number of times that state
        transition has be traversed.

        Args:
            None

        Returns:
            A matplotlib.pyplot that can be titled and shown later (can use plt.title
            or plt.show()

        """
        import networkx as nx
        import matplotlib.pyplot as plt

        d = self.get_property([])

        graph = nx.MultiDiGraph()
        graph.add_edges_from(
            [("Detect", "Polling"), ("Polling", "Configuration"), ("Configuration", "L0")],
            label="1",
        )
        graph.add_edges_from(
            [
                ("Polling", "Detect"),
                ("Configuration", "Detect"),
                ("Configuration", "Loopback"),
                ("Configuration", "R.Lock"),
                ("Configuration", "Disabled"),
                ("L0", "L0s"),
                ("L0", "R.Lock"),
                ("L0", "L1"),
                ("L1", "R.Lock"),
                ("L0s", "R.Lock"),
                ("L0s", "L0"),
                ("R.Lock", "Configuration"),
                ("R.Lock", "R.Cfg"),
                ("R.Lock", "R.Eq"),
                ("R.Lock", "Detect"),
                ("R.Cfg", "R.Speed"),
                ("R.Cfg", "R.Idle"),
                ("R.Cfg", "Configuration"),
                ("R.Cfg", "Detect"),
                ("R.Eq", "R.Lock"),
                ("R.Eq", "R.Speed"),
                ("R.Speed", "Detect"),
                #  ("R.Idle", "R.Lock"),
                ("R.Idle", "Loopback"),
                ("R.Idle", "L0"),
                ("R.Idle", "Detect"),
                ("R.Idle", "Hot reset"),
                ("R.Idle", "Disabled"),
                ("Loopback", "Detect"),
                ("Hot reset", "Detect"),
                ("Disabled", "Detect"),
            ],
            weight=2,
            label="999",
        )

        # visual placement of states
        pos = {
            "Detect": [2, 0],
            "Polling": [2, -2],
            "Configuration": [2, -4],
            "L0": [2, -6],
            "L1": [1, -7],
            "L0s": [2, -8],
            "R.Lock": [3, -8],
            "R.Speed": [3, -10],
            "R.Cfg": [4, -8],
            "R.Idle": [5, -8],
            "R.Eq": [4, -10],
            "Loopback": [4, -6],
            "Hot reset": [4, -4],
            "Disabled": [4, -2],
        }
        plt.figure(figsize=(8, 8))
        font = {"color": "k", "fontweight": "bold", "fontsize": 14}
        plt.title("PCIe LTSSM Graph", font)
        val_map = {}
        for state in pos.keys():
            s = "state." + state.replace(" ", "").lower()
            val_map[state] = "orange" if d[s] == 2 else "green" if d[s] == 1 else "gray"

        values = ["grey" if edge not in val_map else val_map[edge] for edge in graph.nodes()]

        # Specify the edges you want here
        blue_edges = [("R.Lock", "R.Cfg"), ("R.Cfg", "R.Idle"), ("R.Idle", "L0")]
        edge_colors = ["black" if edge not in blue_edges else "blue" for edge in graph.edges()]
        edge_labels = {}
        for e in graph.edges():
            prop_name = "edge." + e[0].lower() + "_" + e[1].lower()
            prop_name = prop_name.replace(" ", "")
            val = d[prop_name]
            edge_labels[e] = val
        nx.draw_networkx_nodes(graph, pos, node_color=values, node_size=1500)

        nx.draw_networkx_labels(graph, pos)
        nx.draw_networkx_edges(graph, pos, edge_color=edge_colors, arrows=True, arrowsize=10)
        nx.draw_networkx_edge_labels(graph, pos, edge_labels=edge_labels)
        return plt



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Union

from chipscopy.api.containers import QueryList
from chipscopy.api.ibert.aliases import (
    ALIAS_DICT,
    CHILDREN,
    GT_KEY,
    HANDLE_NAME,
    MODIFIABLE_ALIASES,
    PLL_KEY,
    PROPERTY_ENDPOINT,
    TYPE,
)
from chipscopy.api.ibert.gt import GT
from chipscopy.api.ibert.pll import PLL
from chipscopy.api.ibert.serial_object_base import SerialObjectBase
from typing_extensions import final

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.api.ibert.ibert import IBERT


@final
class GTGroup(SerialObjectBase["IBERT", Union[GT, PLL]]):
    def __init__(self, name, parent, tcf_node, configuration):
        SerialObjectBase.__init__(
            self,
            name=name,
            type=configuration[TYPE],
            parent=parent,
            handle=configuration[HANDLE_NAME],
            core_tcf_node=tcf_node,
            property_for_alias=configuration.get(ALIAS_DICT, dict()),
            is_property_endpoint=configuration.get(PROPERTY_ENDPOINT, False),
            modifiable_aliases=configuration.get(MODIFIABLE_ALIASES, set()),
        )

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "type": self.type, "handle": self.handle}

        if CHILDREN not in configuration:
            return

        # Build the child objects
        for child_name, child_configuration in configuration[CHILDREN].items():
            if child_configuration[TYPE] == GT_KEY:
                child_class = GT
            elif child_configuration[TYPE].startswith(PLL_KEY):
                child_class = PLL
            else:
                continue

            new_child = child_class(
                name=child_name,
                parent=self,
                tcf_node=self.core_tcf_node,
                configuration=child_configuration,
            )
            self.children.append(new_child)

        self.gts: QueryList[GT] = self.children.filter_by(type=GT_KEY)
        """GT children"""

        self.plls: QueryList[PLL] = self.children.filter_by(type=PLL_KEY)
        """PLL children"""

        # Do this after self.plls has been populated;
        # update_pll() looks through self.plls to determine the PLL for the GT
        for gt in self.gts:
            gt.update_pll()

    def __repr__(self) -> str:
        return self.name

    def reset(self):
        """
        Reset the RXs, TXs and PLLs that are part of the GT Group
        """
        for gt in self.gts:
            gt.reset()
        for pll in self.plls:
            pll.reset()



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from chipscopy.tcf import services
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.proxies.CorePropertyProxy import CorePropertyProxy
from chipscopy.client.core_property_client import PropertyValues

SERVICE_NAME = "AxisDDR"


class AxisDDRProxy(CorePropertyProxy, services.Service):
    def __init__(self, channel):
        super().__init__(channel)
        self._listeners = {}

    #
    #  General methods
    #
    def getName(self):
        return SERVICE_NAME

    def initialize(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("initialize", (node_id,), done)

    def terminate(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("terminate", (node_id,), done)

    def refresh_property(
        self, node_id: str, prop_names: List[str], done: DoneHWCommand
    ) -> PropertyValues:
        return self.send_xicom_command("refreshProperty", (node_id, prop_names), done)

    def refresh_cal_status(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("refreshCalStatus", (node_id,), done)

    def refresh_cal_margin(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("refreshCalMargin", (node_id,), done)

    def get_cal_margin_mode(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("getCalMarginMode", (node_id,), done)



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

import time

# Error report attribute names

ERROR_CODE = "Code"
"""Integer : error code value"""

ERROR_TIME = "Time"
"""Integer : error time"""

ERROR_SERVICE = "Service"
"""String : name of the service the error occured in"""

ERROR_FORMAT = "Format"
"""String : error format"""

ERROR_PARAMS = "Params"
"""List : a list of parameters for the given error"""

ERROR_SEVERITY = "Severity"
"""Integer : error severity"""

ERROR_ALT_CODE = "AltCode"
"""Integer : error alternative code value"""

ERROR_ALT_ORG = "AltOrg"
"""Integer : error alternative org???"""

ERROR_CAUSED_BY = "CausedBy"
"""Object : the cause of the error"""

# Error severity codes
SEVERITY_ERROR = 0
"""Error"""

SEVERITY_WARNING = 1
"""Warning"""

SEVERITY_FATAL = 2
"""Fatal error"""

# Error code ranges
# Standard TCF code range
CODE_STD_MIN = 0
"""Minimum error code value"""
CODE_STD_MAX = 0xffff
"""Maximum error code value"""

# Service specific codes. Decoding requires service ID.
CODE_SERVICE_SPECIFIC_MIN = 0x10000
"""Service specific minimum error code value"""
CODE_SERVICE_SPECIFIC_MAX = 0x1ffff
"""Service specific maximum error code value"""

# Reserved codes - will never be used by the TCF standard
CODE_RESERVED_MIN = 0x20000
"""Service specific reserved minimum error code value"""
CODE_RESERVED_MAX = 0x2ffff
"""Service specific reserved maximum error code value"""

# Standard TCF error codes
TCF_ERROR_OTHER = 1
"""Other TCF errors"""

TCF_ERROR_JSON_SYNTAX = 2
"""JSON syntax error"""

TCF_ERROR_PROTOCOL = 3
"""Protocol error"""

TCF_ERROR_BUFFER_OVERFLOW = 4
"""Buffer overflow"""

TCF_ERROR_CHANNEL_CLOSED = 5
"""Channel closed"""

TCF_ERROR_COMMAND_CANCELLED = 6
"""Command canceled"""

TCF_ERROR_UNKNOWN_PEER = 7
"""Unknown peer"""

TCF_ERROR_BASE64 = 8
"""Base64 error"""

TCF_ERROR_EOF = 9
"""End of file"""

TCF_ERROR_ALREADY_STOPPED = 10
"""Already stopped"""

TCF_ERROR_ALREADY_EXITED = 11
"""Already exited"""

TCF_ERROR_ALREADY_RUNNING = 12
"""Already running"""

TCF_ERROR_ALREADY_ATTACHED = 13
"""Already attached"""

TCF_ERROR_IS_RUNNING = 14
"""Is running"""

TCF_ERROR_INV_DATA_SIZE = 15
"""Invalid data size"""

TCF_ERROR_INV_CONTEXT = 16
"""Invalid context"""

TCF_ERROR_INV_ADDRESS = 17
"""Invalid address"""

TCF_ERROR_INV_EXPRESSION = 18
"""Invalid expression"""

TCF_ERROR_INV_FORMAT = 19
"""Invalid format"""

TCF_ERROR_INV_NUMBER = 20
"""Invalid number"""

TCF_ERROR_INV_DWARF = 21
"""Invalid dwarf"""

TCF_ERROR_SYM_NOT_FOUND = 22
"""Symbol not found"""

TCF_ERROR_UNSUPPORTED = 23
"""Unsupported"""

TCF_ERROR_INV_DATA_TYPE = 24
"""Invalid data type"""

TCF_ERROR_INV_COMMAND = 25
"""Invalid command"""

TCF_ERROR_INV_TRANSPORT = 26
"""Invalid transport"""

TCF_ERROR_CACHE_MISS = 27
"""Cache miss"""

TCF_ERROR_NOT_ACTIVE = 28
"""Not active"""

_timestamp_format = "%Y-%m-%d %H:%M:%S"


class ErrorReport(Exception):
    """TCF error report class.

    :param msg: error report message
    :param attrs: TCF error report attributes to initialise this error report
                  with. See **ERROR_***.
    """
    def __init__(self, msg, attrs):
        super(ErrorReport, self).__init__(msg)
        if isinstance(attrs, int):
            attrs = {
                ERROR_CODE: attrs,
                ERROR_TIME: int(time.time() * 1000),
                ERROR_FORMAT: msg,
                ERROR_SEVERITY: SEVERITY_ERROR
            }
        self.attrs = attrs
        caused_by = attrs.get(ERROR_CAUSED_BY)
        if caused_by:
            errMap = caused_by
            bf = 'TCF error report:\n'
            bf += appendErrorProps(errMap)
            self.caused_by = ErrorReport(bf, errMap)

    def getErrorCode(self):
        """Get this exception error code.

        :returns: This error report error code, or **0**
        """
        return self.attrs.get(ERROR_CODE) or 0

    def getAltCode(self):
        """Get this exception alternative error code.

        :returns: This error report alternative error code, or **0**
        """
        return self.attrs.get(ERROR_ALT_CODE) or 0

    def getAltOrg(self):
        """Get this exception alternative org ???

        :returns: This error report alernative org???, or **None**
        """
        return self.attrs.get(ERROR_ALT_ORG)

    def getAttributes(self):
        """Get this error attribute.

        :returns: a :class:`dict` of this error attributes
        """
        return self.attrs


def toErrorString(data):
    if not data:
        return None
    errMap = data
    fmt = errMap.get(ERROR_FORMAT)
    if fmt:
        c = errMap.get(ERROR_PARAMS)
        if c:
            return fmt.format(c)
        return fmt
    code = errMap.get(ERROR_CODE)
    if code is not None:
        if code == TCF_ERROR_OTHER:
            alt_org = errMap.get(ERROR_ALT_ORG)
            alt_code = errMap.get(ERROR_ALT_CODE)
            if alt_org and alt_code:
                return "%s Error %d" % (alt_org, alt_code)
        return "TCF Error %d" % code
    return "Invalid error report format"


def appendErrorProps(errMap):
    timeVal = errMap.get(ERROR_TIME)
    code = errMap.get(ERROR_CODE)
    service = errMap.get(ERROR_SERVICE)
    severity = errMap.get(ERROR_SEVERITY)
    alt_code = errMap.get(ERROR_ALT_CODE)
    alt_org = errMap.get(ERROR_ALT_ORG)
    bf = ''
    if timeVal:
        bf += '\nTime: '
        bf += time.strftime(_timestamp_format, time.localtime(timeVal / 1000.))
    if severity:
        bf += '\nSeverity: '
        if severity == SEVERITY_ERROR:
            bf += 'Error'
        elif severity == SEVERITY_FATAL:
            bf += 'Fatal'
        elif severity == SEVERITY_WARNING:
            bf += 'Warning'
        else:
            bf += 'Unknown'
    bf += '\nError text: ' + str(toErrorString(errMap)) + '\n'
    bf += 'Error code: ' + str(code)
    if service:
        bf += '\nService: ' + str(service)
    if alt_code:
        bf += '\nAlt code: ' + str(alt_code)
        if alt_org:
            bf += '\nAlt org: ' + str(alt_org)
    return bf



# %% [markdown] pycharm={"name": "#%%\n"}
# <link rel="preconnect" href="https://fonts.gstatic.com">
# <link href="https://fonts.googleapis.com/css2?family=Fira+Code&display=swap" rel="stylesheet">
#
# ### License
#
# <p style="font-family: 'Fira Code', monospace; font-size: 1.2rem">
# Copyright 2021 Xilinx, Inc.<br><br>
# Licensed under the Apache License, Version 2.0 (the "License");<br>
# you may not use this file except in compliance with the License.<br><br>
# You may obtain a copy of the License at <a href="http://www.apache.org/licenses/LICENSE-2.0"?>http://www.apache.org/licenses/LICENSE-2.0</a><br><br>
# Unless required by applicable law or agreed to in writing, software<br>
# distributed under the License is distributed on an "AS IS" BASIS,<br>
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.<br>
# See the License for the specific language governing permissions and<br>
# limitations under the License.<br>
# </p>
#

# %% [markdown]
# # ChipScoPy Memory Read and Write Example
#
#
# <img src="../img/api_overview.png" width="500" align="left">

# %% [markdown]
# ## Description
# This demo shows how to read and write memory in the device using the ChipScoPy Python API.
#
#
# ## Requirements
# - Local or remote Xilinx Versal board, such as a VCK190
# - Xilinx hw_server 2021.2 installed and running
# - Python 3.8 or greater installed
# - ChipScoPy 2021.2 installed
# - Jupyter notebook support installed - Please do so, using the command `pip install chipscopy[jupyter]`

# %% [markdown]
# ## 1 - Initialization: Imports and File Paths
#
# After this step,
# - Required functions and classes are imported
# - Paths to server(s) and files are set correctly

# %% pycharm={"name": "#%%\n"}
import os
from chipscopy import get_design_files
from chipscopy import create_session, report_versions

# %%
# Make sure to start the hw_server prior to running.
# Specify location of the running hw_server below.
# The default is localhost - but can be other locations on the network.
HW_URL = os.getenv("HW_SERVER_URL", "TCP:localhost:3121")
print(f"HW_URL={HW_URL}")

# %% [markdown]
# ## 2 - Create a session and connect to the hw_server
#
# The session is a container that keeps track of devices and debug cores.
# After this step,
# - Session is initialized and connected to server(s)
# - Versions are detected and reported to stdout
#
# *NOTE*: No cs_server is required for this example.

# %% pycharm={"name": "#%%\n"}
session = create_session(hw_server_url=HW_URL)
report_versions(session)

# %% [markdown]
# ## Step 3 - Get the device from the session

# %% pycharm={"name": "#%%\n"}
# Typical case - one device on the board - get it.
versal_device = session.devices.filter_by(family="versal").get()
print(versal_device)

# %% [markdown]
# ## Step 4 - Reset the device

# %% pycharm={"name": "#%%\n"}
versal_device.reset()
print("Reset complete.")

# %% [markdown]
# ## Step 5 - Write and Read memory
#
#
# ChipScoPy can be used to read and write memory using the hardware server.
# Memory reads and writes work similar to xsdb mrd and mwr commands.
#

# %% [markdown]
# ### Show the list of all memory targets
#
# Memory targets in this list can be used for memory_read and memory_write
# operations.

# %% pycharm={"name": "#%%\n"}
print("\nMemory Targets: ", versal_device.memory_target_names)

# %% [markdown]
# ### Simple Write and read memory example
#
# This is the most basic memory_read and memory_write example using the default
# DPC memory target.
#
# Below we write 32-bit values to the specified address and read them back.

# %% pycharm={"name": "#%%\n"}
addr = 0xF2010000
values_to_write = [0x10111213, 0x14151617]

print("\nWriting [{}]".format(", ".join(hex(x) for x in values_to_write)), "to address:", hex(addr))
# Write to the DPC default target
versal_device.memory_write(addr, values_to_write)

print(f"Reading {len(values_to_write)} values from address: hex(addr)")
read_values = versal_device.memory_read(address=addr, num=len(values_to_write))

print("Readback result: [{}]".format(", ".join(hex(x) for x in read_values)))

assert read_values == values_to_write

# %% [markdown]
# ### Changing Memory Read/Write Word Sizes
#
# It is possible to specify the word size when reading and writing.
# Default is 'w'. Other sizes shown below.
# ```
# 'b'=byte
# 'h'=half
# 'w'=word
# 'd'=double word
# ```

# %% pycharm={"name": "#%%\n"}
addr = 0xF2010000
values_to_write = [0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17]

print("\nWriting [{}]".format(", ".join(hex(x) for x in values_to_write)), "to address:", hex(addr))
versal_device.memory_write(addr, values_to_write, size="b")

print("Reading from address: ", hex(addr))
read_values = versal_device.memory_read(address=addr, size="b", num=len(values_to_write))
print("Readback result: [{}]".format(", ".join(hex(x) for x in read_values)))
assert read_values == values_to_write

values_to_write = [0x1000, 0x1234, 0x4321, 0x1313, 0x1414, 0x1515, 0x1616, 0x1717]
print("\nWriting [{}]".format(", ".join(hex(x) for x in values_to_write)), "to address:", hex(addr))
versal_device.memory_write(addr, values_to_write, size="h")

print("Reading from address: ", hex(addr))
read_values = versal_device.memory_read(address=addr, size="h", num=len(values_to_write))
print("Readback result: [{}]".format(", ".join(hex(x) for x in read_values)))
assert read_values == values_to_write

# %% [markdown]
# ### Selecting different memory targets and improving performance
#
# The examples above use the device class for memory_read() and memory_write()
# operations. Using the device read and write is simple, but has additional
# overhead with each call to find the associated memory context.
#
# It is possible to explicitly request the memory context for a desired target.
# Once a memory context is obtained, memory_read and memory_write operations
# can be executed on that target repeatedly.
#
# This eliminate some of the additional overhead.
#
# The example below shows how to get a context to repeatedly read and write from
# different memory targets.

# %% pycharm={"name": "#%%\n"}
addr = 0xF2010000
dpc = versal_device.memory.get(name="DPC")
apu = versal_device.memory.get(name="APU")
for i in range(10):
    values_to_write = [0x12345678 + i, 0xFEDCBA98 - i]
    # Write to the DPC without context lookup overhead
    print(
        "\nDPC: Writing [{}]".format(", ".join(hex(x) for x in values_to_write)),
        "to address:",
        hex(addr),
    )
    dpc.memory_write(addr, values_to_write)
    # Read from the APU without context lookup overhead
    print("APU: Reading from address: ", hex(addr))
    read_values = apu.memory_read(address=addr, num=len(values_to_write))
    print("Readback result: [{}]".format(", ".join(hex(x) for x in read_values)))
    assert read_values == values_to_write



# %% [markdown]
# <link rel="preconnect" href="https://fonts.gstatic.com">
# <link href="https://fonts.googleapis.com/css2?family=Fira+Code&display=swap" rel="stylesheet">
#
# ### License
#
# <p style="font-family: 'Fira Code', monospace; font-size: 1.2rem">
# Copyright 2021 Xilinx, Inc.<br><br>
# Licensed under the Apache License, Version 2.0 (the "License");<br>
# you may not use this file except in compliance with the License.<br><br>
# You may obtain a copy of the License at <a href="http://www.apache.org/licenses/LICENSE-2.0"?>http://www.apache.org/licenses/LICENSE-2.0</a><br><br>
# Unless required by applicable law or agreed to in writing, software<br>
# distributed under the License is distributed on an "AS IS" BASIS,<br>
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.<br>
# See the License for the specific language governing permissions and<br>
# limitations under the License.<br>
# </p>
#

# %% [markdown]
# # ChipScoPy SPTG Perfmon Example
#
#
# <img src="../img/api_overview.png" width="500" align="left">

# %% [markdown]
# ## Description
# This example demonstrates how to configure a Versal for taking NoC performance measurements.
#
# ## Requirements
# - Local or remote Xilinx Versal board, such as a VCK190
# - Xilinx hw_server 2021.2 installed and running
# - Xilinx cs_server 2021.2 installed and running
# - Python 3.8 or greater installed
# - ChipScoPy 2021.2 installed
# - Jupyter notebook support installed - Please do so, using the command `pip install chipscopy[jupyter]`
# - Matplotlib support installed - Please do so, using the command `pip install chipscopy[core-addons]`

# %% [markdown]
# ## 1 - Initialization: Imports and File Paths
#
# After this step,
# - Required functions and classes are imported
# - Paths to server(s) and files are set correctly

# %%
import os
from time import sleep
import matplotlib  # for nbconvert'd script
from chipscopy.api.noc import (
    TC_BEW,
    TC_BER,
    NoCPerfMonNodeListener,
    PerfTGController,
)
from chipscopy.api.noc.plotting_utils import MeasurementPlot
from chipscopy import create_session, report_versions
from chipscopy import get_design_files

# %%
# Make sure to start the hw_server and cs_server prior to running.
# Specify locations of the running hw_server and cs_server below.
# The default is localhost - but can be other locations on the network.
CS_URL = os.getenv("CS_SERVER_URL", "TCP:localhost:3042")
HW_URL = os.getenv("HW_SERVER_URL", "TCP:localhost:3121")

# The get_design_files() function tries to find the PDI and LTX files. In non-standard
# configurations, you can put the path for PROGRAMMING_FILE and PROBES_FILE below.
design_files = get_design_files("vck190/production/chipscopy_ced")

PROGRAMMING_FILE = design_files.programming_file
PROBES_FILE = design_files.probes_file

print(f"HW_URL: {HW_URL}")
print(f"CS_URL: {CS_URL}")
print(f"PROGRAMMING_FILE: {PROGRAMMING_FILE}")
print(f"PROBES_FILE:{PROBES_FILE}")

# %% [markdown]
# ## 2 - Create a session and connect to the hw_server and cs_server
#
# The session is a container that keeps track of devices and debug cores.
# After this step,
# - Session is initialized and connected to server(s)
# - Versions are detected and reported to stdout

# %%
session = create_session(cs_server_url=CS_URL, hw_server_url=HW_URL)
report_versions(session)

# %% [markdown]
# ## 3 - Program the device with the example design
#
# After this step,
# - Device is programmed with the example programming file

# %%
versal_device = session.devices.filter_by(family="versal").get()
versal_device.program(PROGRAMMING_FILE)

# %% [markdown]
# ## 4 - Discover Debug Cores
#
# Debug core discovery initializes the chipscope server debug cores. This brings debug cores in the chipscope server online.
#
# After this step,
#
# - The cs_server is initialized and ready for use

# %%
versal_device.discover_and_setup_cores(noc_scan=True, ltx_file=PROBES_FILE)
print(f"Debug cores setup and ready for use.")

# %% [markdown]
# ## 5 - Setup NoC core
#
# Ensure scan nodes are enabled in the design.

# %%
# We begin by enumerating the debug cores (hard and soft) present in the design.
# Then we ask the design for the supported timebases. And, finally:
# The NoC is scanned to determine the activated elements.

noc = versal_device.noc_core.get()

scan_nodes = ["DDRMC_X0Y0", "NOC_NMU512_X0Y0"]
print("\nEnumerating nodes: ", end="")
for node in scan_nodes:
    print(f"{node}, ", end="")
print("...", end="")

# this will setup the nodes on the server side and return the nodes successfully enumerated
enable_list = noc.enumerate_noc_elements(scan_nodes)
print("complete!")

# %%
supported_periods = noc.get_supported_sampling_periods(100 / 3, 100 / 3, {"DDRMC_X0Y0": 800.0})
print("Supported sampling periods:")
for domain, periods in supported_periods.items():
    print(f"  {domain}:")
    for p in periods:
        print(f"    {p:.0f}ms", end="")
    print()

# %%
# Select Timebase and Nodes to Monitor
#
# For the two clock domains we must select a sampling period from the hardware supported values. The debug cable used will dictate how much bandwidth is available, so high frequency sampling may not actually produce data at the specified rate. Recommendation is ~500ms for jtag.
#
# Then the user must decide what to monitor--again the bandwidth is a definite consideration here. Plot performance may become the bottleneck (Optimizations will come later in the renderer or agg backend). The guidance here is to pick up to 4 nodes to monitor.

desired_period = 500  # ms
sampling_intervals = {}

for domain in supported_periods.keys():
    sampling_intervals[domain] = 0
    for sp in supported_periods[domain]:
        if sp > desired_period:
            sampling_intervals[domain] = sp
            break

    if sampling_intervals[domain] == 0:
        print(
            f"Warning, desired period {desired_period}ms is slower than "
            f"longest supported period {supported_periods[domain][-1]}ms [{domain} domain] "
            f"defaulting to this value."
        )
        sampling_intervals[domain] = supported_periods[-1]

print(f"Sampling period selection:")
for domain, freq in sampling_intervals.items():
    print(f"  {domain}: {freq:.0f}ms")

# %%
# Configure Monitors
#
# As a precaution, it's a good idea to validate the desired nodes are enabled for the design.
# Using the set of nodes and the desired sampling periods it's time to ask the service to start pushing metric data. Two additional arguments are required to this API.
#
# Traffic class
#
# This is a bit-or field of the requested traffic classes. Note, One monitor is dedicated to read traffic classes and the other to write--so all read TCs will apply to one channel and all write TCs to the other. `Best effort` is a good place to start.
#
# Number of Samples
#
# The number of samples allows for a burst of measurements to be taken and then the underlying service will tear down the monitors and stop pumping data back to the client. `-1` denotes that sampling shall continue indefinitely.

# total number of samples to capture (-1 for continuous mode)
num_samples = -1

print("Setting up monitors for: ")
for node in enable_list:
    print(node)

# When overflow occurs the precision of the monitors must be traded for range
# See the server API for more information
extended_monitor_config = {"NOC_NMU512_X0Y0": {"tslide": 0x3}}  # or None
noc.configure_monitors(
    enable_list, sampling_intervals, (TC_BEW | TC_BER), num_samples, None, extended_monitor_config
)


# %% [markdown]
# ## 6 - Traffic Generator Configuration
#
# Release the traffic generator from reset.

# %%
vio_cores = versal_device.vio_cores
for index, vio_core in enumerate(vio_cores):
    print(f"\nVIO Core Index {index}")
    print("NAME       :", vio_core.name)
    print("UUID       :", vio_core.uuid)
    print("PORT_NAMES :", vio_core.port_names)
    print("PROBE_NAMES:", vio_core.probe_names)

# tg_vio_bc = "sptg_axi_i"
tg_vio_bc = "chipscopy_i/noc_tg_bc"
tg_vio_name = f"{tg_vio_bc}/axis_vio_0"
print(tg_vio_name)
tg_vio = versal_device.vio_cores.get(name=tg_vio_name)

tg_vio.reset_vio()
tg_vio.write_probes(
    {f"{tg_vio_bc}/noc_sim_trig_rst_n": 0x1, f"{tg_vio_bc}/noc_tg_tg_rst_n": 0x1}
)  # clear nrst on trigger and tg

first_tg_ba = 0x201_8000_0000
tg = PerfTGController(first_tg_ba, versal_device, vio=tg_vio)

# %% [markdown]
# ## 7 - Create plotter and listener
#
# Attach both to running view

# %%
record_to_file = False  # True | False
node_listener = NoCPerfMonNodeListener(
    sampling_intervals,
    num_samples,
    enable_list,
    record_to_file,
    extended_monitor_config=extended_monitor_config,
)
session.chipscope_view.add_node_listener(node_listener)

plotter = MeasurementPlot(enable_list, mock=False, figsize=(10, 7.5), tg=tg)
node_listener.link_plotter(plotter)

# Build Plotting Graphs
matplotlib.use("Qt5Agg")
plotter.build_graphs()

# %% [markdown]
# ## 8 - Main Event Loop
#
# This loop runs until you close the plotter.
# If you are using a finite amount of measurement samples, you can uncomment the if --> break statement to automatically return from execution of this cell upon completion of the burst.

# %%
# Run Main Event Loop
loop_count = 0
while True:
    session.chipscope_view.run_events()
    sleep(0.1)
    plotter.fig.canvas.draw()
    plotter.fig.canvas.flush_events()
    if not plotter.alive:
        break
    # Below will return on burst completion - uncomment if you want to try.
    # if all([x <= 0 for x in node_listener.num_samples.values()]):
    #     break

# %%
# Reset Traffic Generator
# This allows for a hard block-level reset of the traffic generator.

tg.block_reset()



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import copy
from typing import List, Union, ClassVar
import logging
from pathlib import Path
from logging import handlers

import loguru


class ParodyLogger:
    def trace(self, *args, **kwargs):
        pass

    def debug(self, *args, **kwargs):
        pass

    def info(self, *args, **kwargs):
        pass

    def warning(self, *args, **kwargs):
        pass

    def success(self, *args, **kwargs):
        pass

    def error(self, *args, **kwargs):
        pass

    def critical(self, *args, **kwargs):
        pass

    def __getattr__(self, item):
        return self

    def __getitem__(self, item):
        return self


class CSSLogger:
    # Default format
    default_format: ClassVar[str] = (
        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
        "<level>{level: <8}</level> | "
        "<m>{extra[domain]: >20}</m> | "
        "<cyan>{name: >50}</cyan> | "
        "<cyan>{line: <4}</cyan> | "
        "<level>{message}</level>"
    )

    def __init__(self):
        self._parody_logger = ParodyLogger()

        loguru.logger.remove()

        self.logger: loguru.Logger = copy.deepcopy(loguru.logger)
        # Enable all modules, we'll enforce domain using record filter
        self.logger.enable("")

        self.logger_for_domain = dict()

        self.domain_enabled = dict()

        self.default_sink_id: int = None

        self.current_log_level: loguru.Level = None

    def __call__(self, *args, **kwargs):
        if len(args) == 0:
            return
        msg = args[0]
        log["client"].info(msg)

    def __getattr__(self, item) -> loguru.logger:
        if not self.domain_enabled.get(item, False):
            return self._parody_logger
        return self._get_logger_for_domain(item)

    def __getitem__(self, item) -> loguru.logger:
        if not self.domain_enabled.get(item, False):
            return self._parody_logger
        return self._get_logger_for_domain(item)

    def _send_record_to_sink(self, record) -> bool:
        return self.domain_enabled.get(record["extra"]["domain"], False)

    def _get_logger_for_domain(self, domain):
        if domain not in self.logger_for_domain:
            new_logger = self.logger.bind(domain=domain)
            self.logger_for_domain[domain] = new_logger
            if domain not in self.domain_enabled:
                # Disable domain logging by default.
                self.domain_enabled[domain] = False

        return self.logger_for_domain[domain]

    def is_domain_enabled(self, domain: str, level: str) -> bool:
        level_info = self.logger.level(level)
        return self.domain_enabled.get(domain, False) and level_info.no >= self.current_log_level.no

    def enable_domain(self, domain_name: Union[str, List[str]]):
        if isinstance(domain_name, str):
            domain_name = [domain_name]

        for domain in domain_name:
            # Need not check if domain is in domain_logging_status.
            # This is in case the user enables the domain before logging any message.
            self.domain_enabled[domain] = True

    def disable_domain(self, domain_name: Union[str, List[str]]):
        if isinstance(domain_name, str):
            domain_name = [domain_name]

        for domain in domain_name:
            if domain in self.domain_enabled:
                self.domain_enabled[domain] = False

    def change_log_level(self, level: str):
        level_info = self.logger.level(level)

        if self.default_sink_id is not None:
            # Delete only the stdout sink. Don't touch any others that might have been added by user
            self.logger.remove(self.default_sink_id)

        self.default_sink_id = self.logger.add(
            sink=sys.stdout,
            level=level,
            format=CSSLogger.default_format,
            filter=self._send_record_to_sink,
        )

        self.current_log_level = level_info

    def add_file_handler(self, full_path: Union[str, Path], level_name: str) -> int:
        level_info = self.logger.level(level_name)
        return self.logger.add(full_path, format=CSSLogger.default_format, level=level_info.name)

    def add_queue_handler(self, queue, level_name: str) -> int:
        queue_handler = logging.handlers.QueueHandler(queue)
        level_info = self.logger.level(level_name)
        return self.logger.add(
            queue_handler, format=CSSLogger.default_format, level=level_info.name
        )


# NOTE - Server side code should not import this! Only client side code should use this.
#  Server side code should import from init in server
log = CSSLogger()
